unit DUTUtilsFiles;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   22 June 1999
   Updated 25 April 2001 (Serious bug in GetFileNames fixed)
   Updated 09 February 2002 (Resource protection tightened and a small bug fixed in DUTGetUserDomain)
   Updated 02 April 2002 (Fixed serious bug in trash methods)
           24 May 2002 (Fixed minor bug with trash methods and Win 2k)
    1999-2002 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   File name and file system utilities.
*)
interface

uses
   Classes,  { TStringList }
   Windows;  { TLargeInteger, MAX_PATH, etc. }

const
   sBackslash = '\';
   DriveLetterSet = ['A'..'Z', 'a'..'z'];

type
   TDriveLetter = 'A'..'Z';

   TSpecialFolderType = (sfAppData,        // Windows\Profiles\<user name>\Application Data
                         sfCommonAppData,   // WinNT: All Users\Application Data - Win95: Windows\Profiles\<user name>\Application Data
                         sfCommonDesktop,   // WinNT: All Users\Desktop          - Win95: Windows\Desktop
                         sfCommonPrograms,  // WinNT: All Users\Programs         - Win95: Start Menu\Programs
                         sfCommonStartMenu, // WinNT: All Users\Start Menu       - Win95: Windows\[Profiles\<user name>\]Start Menu
                         sfCommonStartup,   // WinNT: All Users\Startup          - Win95: Start Menu\Programs\Startup
                         sfCommonFavorites, // WinNT: All Users\Favorites        - Win95: Windows\Favorites
                         sfUserDesktop,     // Windows\[Profiles\<user name>\]Desktop
                         sfFavorites,       // Windows\[Profiles\<user name>\]Favorites
                         sfFonts,           // Windows\Fonts
                         sfNetworkNeighborhood, // Windows\[Profiles\<user name>\]NetHood
                         sfPersonal,        // My Documents
                         sfPrintersNeighborhood, // Windows\PrintHood
                         sfPrograms,        // Start Menu\Programs
                         sfRecent,          // Windows\[Profiles\<user name>\]Recent
                         sfSendTo,          // Windows\[Profiles\<user name>\]SendTo
                         sfStartMenu,       // Windows\[Profiles\<user name>\]Start Menu
                         sfStartUp,         // Start Menu\Programs\Startup
                         sfTemplates,       // Windows\ShellNew
                         sfInternetCache,   // Windows\Temporary Internet Files
                         sfCookies,         // Windows\Profiles\<user name>\Cookies
                         sfHistory,         // Windows\Profiles\<user name>\History
//                       sfProgramFiles,    // Program Files
//                       sfPictures,        // My Pictures
                         sfWindows,         // Windows
                         sfSystem           // Windows\System
                        );
// TFileName = String;

   TObjOption = (ooIgnore, //Ignore TStringList.Objects[] (will be Nil)
                 ooFileSize, //Integer(TStringList.Objects[]) are file sizes
                 ooFileAttr, //Integer(TStringList.Objects[]) are file attributes
                 ooFileAge   //Integer(TStringList.Objects[]) are file ages (as per the FileAge function (q.v.))
                );           //TDateTime := FileDateToDateTime(Integer(TStringList.Objects[]));

procedure OLECheckBoolean(Result : Boolean);
{
   Transforms a boolean Result into an HResult-like
   value and passes it on to OLECheck.
}

{ #################### Disk Manipulation #################### }

function GetDiskFreeSpaceExA(lpDirectoryName : PAnsiChar; var lpFreeBytesAvailableToCaller, lpTotalNumberOfBytes : TLargeInteger; lpTotalNumberOfFreeBytes : PLargeInteger) : BOOL; stdcall;
function GetDiskFreeSpaceExW(lpDirectoryName : PWideChar; var lpFreeBytesAvailableToCaller, lpTotalNumberOfBytes : TLargeInteger; lpTotalNumberOfFreeBytes : PLargeInteger) : BOOL; stdcall;
function GetDiskFreeSpaceEx (lpDirectoryName : PChar;     var lpFreeBytesAvailableToCaller, lpTotalNumberOfBytes : TLargeInteger; lpTotalNumberOfFreeBytes : PLargeInteger) : BOOL; stdcall;
{
   Replacement declarations for the incorrect ones in Windows.pas
}

{ #################### FileName Manipulation #################### }

//function MinimizeFileName(const FileName : TFileName; Canvas : TCanvas; MaxLen : Integer) : TFileName;
{
   Minimizes the FileName so it'll fit in Width pixels on the Canvas.
   See DUTUtilsGraphics.
}

function HasTrailer(const FileName : String) : Boolean;
{
   Returns True if FileName ends with a '\', False otherwise.
}

function AddTrailer(var FileName : String) : Boolean;
{
   AddTrailer adds the closing '\' (if missing) to FileName.
   Returns True if it was missing, False otherwise.
   An empty FileName is left unchanged (and returns False).
}

function RemoveTrailer(var FileName : String) : Boolean;
{
   RemoveTrailer removes the closing '\' (if any) from FileName.
   Returns True if there was one, False otherwise.
}

function IsDrive(const FileName : String) : Boolean;
{
   Returns True if FileName begins with <DriveLetter> + ':', False otherwise.
}

function IsUNCServer(const FileName : String) : Boolean;
{
   Returns True if FileName begins with '\\', False otherwise.
}

function ExtractFileTitle(const FileName : String) : String;
{
   ExtractFileTitle extracts the title part of the given
   filename.  The resulting string is the rightmost characters of FileName,
   starting with the first character after the colon or backslash that
   separates the path information from the name and stopping short of the
   period and extension.
   The resulting string is empty if FileName contains no title.
}

function ExtractUNCServer(const FileName: string): string;
{
   ExtractUNCServer extracts the UNC server part of the given filename.
   For filenames with a UNC path, the resulting string is in the form
   '\\<servername>'.
   Otherwise, the result is an empty string.
}

function ExtractUNCResource(const FileName: string): string;
{
   ExtractUNCResource extracts the UNC resource part of the given filename.
   For filenames with a UNC path, the resulting string is in the form
   '\<resourcename>'.
   Otherwise, the result is an empty string.
}

procedure ParseFileName(const FileName : String; var Drive : String; var Path : String; var Title : String; var Extension : String);
{
   Parses the FileName into its components (FileName need not exist).
   Drive will be either '', '<drive>:\' or '\\<server>\<resource>\'.
   Path will be either '', '<dir>\<dir>\<...>\' or a relative path.
   Title and Extension make up the file's name per se;
   Extension includes the period.  Either could be null strings.
   If FileName is a directory, behaviour depends on the closing '\'.

   Compare with Delphi's ProcessPath?
}

{ #################### File System Functions #################### }

function FileTimeToDateTime(FileTime : TFileTime; ToLocal : Boolean) : TDateTime;
{
   Converts the FileTime to a TDateTime value.
   When ToLocal is True, converts the FileTime from UTC to Local Time first.
   Returns -1 in case of failure.
}

function IsExtantFileContainer(const FileName : String) : Boolean;
{
   Returns True if FileName is a pure local drive ('<drive>:') or
   UNC drive ('\\<Server>[\<Resource>]') that exists.
   Inaccessible UNC servers and/or resources do not "exist".
   The function is not case-sensitive.
}

function GetLongFileName(FileName : String) : String;
{
   Obtains the Long version of a file name.
   Input must be fully qualified;
   Output will be.
   Returns an empty string if the file does not exist or is inaccessible.
}

function GetShortFileName(FileName : String) : String;
{
   Obtains the Short (8.3) version of a file name.
   Input must be fully qualified;
   Output will be.
   Returns an empty string if the file does not exist or is inaccessible.
}

function GetRemoteName(LocalName : String) : String;
{
   Returns the UNC path and title of the fully qualified LocalName.
   Returns an empty string if the local file isn't shared across
   the network or some other network-error occurs.
   Also works with remote files that have been mapped to a local drive letter.
   The returned path and title are in Long File Name format even
   if the input LocalName was in 8.3 format.
   Example:
   If your machine shares its C:\Temp directory as \\Avior\Temporary, then
   GetRemoteName('C:\Temp\Dummy.bak') returns '\\Avior\Temporary\Dummy.bak'.
   If you have mapped the remote resource '\\Pascal\Temporary\' to 'K:\', then
   GetRemoteName('K:\Dummy.bak') returns '\\Pascal\Temporary\Dummy.bak'.
}

function GetLocalName(UNCName : String) : String;
{
   Returns the local path and title of the UNCName.
   Returns an empty string if the file isn't truly local or if
   the remote resource has not been mapped to a local file.
   The returned path and title are in Long File Name format even
   if the input UNCName was in 8.3 format.
   Example:
   If your machine shares its C:\Temp directory as \\Avior\Temporary, then
   GetLocalName('\\Avior\Temporary\Dummy.bak') returns 'C:\Temp\Dummy.bak'.
   If you have mapped the remote resource '\\Pascal\Temporary\' to 'K:\', then
   GetLocalName('\\Pascal\Temporary\Dummy.bak') returns 'K:\Dummy.bak'.
}

function DirExists(const Dir : String) : Boolean;
{
   Determines whether the Dir exists.
   Dir can be a relative path.
   If Dir is a *file* or an empty string, returns False.
   Works equally well with and without the closing '\'.
   The function is not case-sensitive.
}

function DriveExists(const Drive : String) : Boolean;
{
   Returns True if the Drive exists, False otherwise.
   The function is not case-sensitive.
}

function UNCResourceExists(const ResourceName : String) : Boolean;
{
   Returns True if the ResourceName exists on the network.
   ResourceName should be in the form "\\Server\Resource"
   (any appended path will be ignored).
   The function is not case-sensitive.
}

function UNCServerExists(const ServerName : String) : Boolean;
{
   Returns True if the ServerName exists, False otherwise.
   ServerName should be of the form "\\Name"
   (an appended resource will be used to speed up the network request,
   any appended path will be ignored).
   The function is not case-sensitive.
   There may be a substantial delay if ServerName does not exist (or is inaccessible).
}

function UNCServerConnected(const ServerName : String; Live : Boolean) : Boolean;
{
   Returns True if the UNC server ServerName is currently connected through the network.
   When Live is True, only live connections are checked; otherwise all persistent
   connections are checked.
   ServerName must be of the form "\\Name".
   The function is not case-sensitive.
}

function RemoteComputerExists(const ComputerName : String) : Boolean;
{
   Returns True if the remote computer ComputerName exists on the network.
   ComputerName must be of the form "\\Name".
   The function is not case-sensitive.
   Be warned that this may take a *long* time to return.
   (Much longer than UNCServerExists)
}
(*
function GetRemoteComputers : TStringList;
{
   Returns a TStringList (possibly empty) giving the remote computer names on the network.
   The names are all of the form "\\Name".  The TStringList is not sorted.
   Be warned that this may take a *long* time to return.
}
*)
function IsALink(FileName : String) : Boolean;
{
   Returns True if the FileName is an LNK file.
   The function is not case-sensitive.
}

function DUTGetComputerName : String;
{
   Obtains the local machine's name.
}

function DUTGetComputerServerName : String;
{
   Obtains the local machine's network server name.
   The result does not include the leading '\\'.
}

procedure DUTGetSharedResourceNames(var Resources, Paths : TStrings);
{
   Resources is Cleared and then loaded with the names
   of the UNC resources shared by the local computer,
   prepended with the local computer's UNC server name.
   Resources is also Sorted.
   Paths is likewise Cleared and then loaded with the
   local paths of each resource.
   Both sets of strings include the closing backslashes.
   IMPORTANT: Make sure both TStrings' Sorted property are False.
   Example:
   Resources[0] == '\\Avior\Temporary\'    Paths[0] == 'C:\Temp\'
}

function DUTGetSecurityProvider : String;
{
   Obtains the local machine's network security provider container name.
   This may be different from the name of the domain (workgroup) the
   local machine belongs to (see DUTGetUserDomain).
}

function DUTGetUserDomain : String;
{
   Obtains the domain the computer belongs to.
   This is slow!
}

function DUTGetFileSize(const FileName: string) : Integer;
{
   Returns the FileName's size, in bytes.
   -1 if it does not exist (or is inaccessible), 0 if empty or a directory.
   Huge files (whose size exceeds $7FFF FFFF bytes)
   will raise an EIntOverflow exception.
}

procedure DUTGetDiskFreeSpace(const DriveName: string; var SectorsPerCluster, BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters : Integer);
{
   Returns the DriveName's vital space statistics.
   DriveName must be either a drive (X:\) or an UNC resource name
   (\\Server\Resource; pure UNC servers will fail).
   Normally, UNC paths would fail, but the function circumvents this
   by temporarily mapping the remote resource to a local drive letter.
   The function is not case-sensitive.
}

function GetFileFootprint(const FileName: string) : Integer;
{
   Returns the FileName's footprint, in bytes.
   FileName must be a fully qualified path and title.
   The footprint will be a whole multiple of the underlying drive's cluster size.
   -1 if the file does not exist or is inaccessible.
   Directories and drives (local or remote) report a zero footprint.
   The function is not case-sensitive.
}

function CanFileFit(const FileName, Drive : string) : Boolean;
{
   Whether the FileName can fit on Drive.
   The decision is based on the footprint, not the byte-size.
   FileName must be a fully qualified path and title.
   Drive may be a local drive ('X:\') or a UNC server and resource
   (it'll be temporarily mapped to a drive letter if need be).
   Returns False if the file does not exist or is inaccessible.
   The function is not case-sensitive.
}

function MapDrive(const DriveName : String; ExtantOnly : Boolean; var Letter : TDriveLetter) : Boolean;
{
   Sets Letter to the DriveName's mapped drive letter.
   DriveName must be an UNC resource name (\\Server\Resource; pure UNC servers will fail).
   When ExtantOnly is True, MapDrive sets Letter to the pre-existing mapping.
   If there is none, MapDrive fails.
   When ExtantOnly is False, MapDrive sets Letter to the pre-existing mapping.
   If there is none, MapDrive grabs the Letter passed in if it is free.
   If that also fails (the Letter is already taken), it works its way back up the
   alphabet (from Z towards A) until it finds a free Letter or runs out of letters.
   MapDrive returns True if successfull, False otherwise.
   The function is not case-sensitive.
}

function DUTGetWindowsDirectory : String;
{
   Obtains the Windows directory path.
   The ending backslash is included.
}

function DUTGetSystemDirectory : String;
{
   Obtains the Windows System directory path.
   The ending backslash is included.
}

function DUTGetCurrentDirectory : String;
{
   Obtains the current directory path.
   The ending backslash is included.
}

function DUTGetTempDirectory : String;
{
   Obtains the Temporary directory path.
   The ending backslash is included.
}

function DUTGetTempFilename(Path, Prefix : String) : String;
{
   Obtains a Temporary filename.
   The file exists upon return; it is created in Path (or the
   current temporary directory, if not specified).
   If Prefix is non-blank, the first three ANSI characters
   are used to begin the file title.
   The fully qualified path of the temporary file is returned.
}

function GetSpecialFolderPath(SpecialFolderType : TSpecialFolderType) : String;
{
   Obtains from the Shell the file path to one of the desktop's special folders.
   sfAppData, sfCommonAppData, sfCommonDesktop, sfCommonPrograms,
   sfCommonStartMenu, sfCommonStartup, sfCommonFavorites, sfUserDesktop,
   sfFavorites, sfFonts, sfNetworkNeighborhood, sfPersonal,
   sfPrintersNeighborhood, sfPrograms, sfRecent, sfSendTo, sfStartMenu,
   sfStartUp, sfTemplates, sfInternetCache, sfCookies, sfHistory,
   sfWindows, sfSystem.
   See also GetSpecialFolderLocation in DUTShellLink.pas
   The ending backslash is included.
}

function FileProperties(const FileName : String) : Boolean;
{
   "Shells out" to the Properties window of the specified file or folder.
   Returns True if successfull, False otherwise.
}

function GetFileVersion(const FileName : String) : String;
{
   Obtains the specified file's Version string.
   Returns an empty string if the specified file
   isn't a Win32 executable (or DLL, etc.) or if it doesn't
   include a version string.
   For an executable to obtain it's own version info,
   it should pass in ParamStr(0).
}

function GetFileNames(const Path : String; AttrReq, AttrExc : Integer; ExcludePath, Recurse : Boolean; ObjOption : TObjOption) : TStringList;
{
   Creates a sorted TStringList loaded with the files (and/or directories)
   found at Path (a fully qualified search pattern such as 'C:\*.*')
   Any directories returned will include the closing backslash.
   AttrReq specifies the attributes which *must* be satisfied,
   whilst AttrExc specifies those that must *not* appear.
   Setting both to zero is equivalent to asking for faAnyFile.
   If ExcludePath is set, the file names will exclude the Path part.
   Sub-directories are scanned only if Recurse is True (in which case,
   if ExcludePath is set *relative* file names will be returned).
   ObjOption indicates what to do with the TStringList's Object values.
      In all cases, the TSringList.Objects[] need to be typecast as Integer.
      ooIgnore   - Nothing.  All Object values will be Nil.
      ooFileSize - Object values are file sizes
      ooFileAttr - Object values are file attributes
      ooFileAge  - Object values are file ages (as per the FileAge function (q.v.))
                   TDateTime := FileDateToDateTime(Integer(TStringList.Objects[i]));
}

{ #################### Trash (Recycle Bin) Functions #################### }

function GetTrashContents(DeletionDateTimes, FileSizes : TList) : TStringList;
{
   Returns a TStringList loaded with the fully qualified names of the files
   currently in the Trash (a.k.a Recycle Bin).
   The TStringList could be empty.
   It is the responsability of the calling application to Free the TStringList.
   If either or both of DeletionDateTimes and FileSizes are non-Nil, they will be
   loaded (in the same sequence as the TStringList) with the files' deletion
   times (as pointers to TDateTime objects) and sizes (as pointers to Integers).
   It is the responsability of the calling application to Free the pointed
   TDateTime and/or Integer objects.
}

function RestoreFromTrash(FileName : String) : Boolean;
{
   Attempts to Restore the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}

function EmptyTrash : Boolean;
{
   Empties the Trash (a.k.a Recycle Bin).
   Returns False if the trash was already empty.
}

implementation

uses
{$IFDEF Debug}
Dialogs, { MessageDlg }
{$ENDIF}
   SysUtils, { ANSIPos }
   ShlObj,   { CSIDL_, SHGetSpecialFolderLocation, etc. }
   ShlObjAdditional,
   ShellAPI, { TShellExecuteInfo }
   Forms,    { TApplication }
   Registry, { TRegistry }
   RegStr,   { REGSTR_... }
   COMObj,   { OLECheck }
   ActiveX,  { IMalloc }
// DUTPointer,   { TPointer, TString }
   DUTUtilsMisc, { FlagIncluded }
   DUTUtilsSys;  { IsWindowsNT }

const
   //Strings that may need to be localised
   RecycledDirStr   = 'Recycled';
   RecycledDirNTStr = 'Recycler'; //A strange subdir also appends itself...
   RestoreCmdStr    = 'R&estore';
   EmptyTrashCmdStr = 'Empty Recycle &Bin';
   //from LMErr.h
   NERR_SUCCESS     = 0;

   //Language independent string
   CMDSTR_DELETE    = 'delete';

procedure OLECheckBoolean(Result : Boolean);
{
   Transforms a boolean Result into an HResult-like
   value and passes it on to OLECheck.
}
begin
   OLECheck(-Ord(not Result));
end; { OLECheckBoolean }

{ #################### Private Support Functions #################### }

function NetEnumerate(lpnr : PNetResource; Scope, ResourceType : DWORD; const ComputerName : String) : Boolean;
{
   Private network enumeration recursive function used by UNCServerConnected and RemoteComputerExists.
   lpnr should initially be Nil to start the enumeration at the network's root.
   Scope should be RESOURCE_GLOBALNET, RESOURCE_CONNECTED or RESOURCE_REMEMBERED.
   ResourceType should be RESOURCETYPE_ANY, RESOURCETYPE_DISK or RESOURCETYPE_PRINT.
}
var
   hEnum        : THandle;
   BufferSize,
   NumEntries,
   Entry        : Integer;
   lpnrLocalc,  //Actually a pointer to an Array of TNetResource
   P            : PNetResource;
begin
   Result := False;
   if NO_ERROR <> WNetOpenEnum(Scope, ResourceType, 0 { Usage: All resources }, lpnr, hEnum) then begin
      Exit;
   end; { then }
   BufferSize := 16384; { 16 kB }
   lpnrLocalc := AllocMem(BufferSize);
   repeat
      { Get as many entries as possible; NumEntries will be set to the number actually read (if successfull) }
      NumEntries := $FFFFFFFF;
      case WNetEnumResource(hEnum, NumEntries, lpnrLocalc, BufferSize) of
      NO_ERROR : begin
         P := lpnrLocalc;
         for Entry := 0 to (NumEntries - 1) do begin
            { P^.dwScope will be Scope since that is what we asked for }
            {
            Compare the lpRemoteName up to ComputerName's length;
            since lpRemoteName always includes a Resource part,
            we must do this to correctly recognise Servers
            }
            if 0 = ANSICompareText(Copy(P^.lpRemoteName, 1, Length(ComputerName)), ComputerName) then begin
               { P^.dwDisplayType = RESOURCEDISPLAYTYPE_SERVER should also be True }
               { FlagIncluded(P^.dwUsage, RESOURCEUSAGE_CONTAINER) should also be True }
               Result := True;
               break;
            end; { then }
            { ResourceType is irrelevant }

            if P^.dwDisplayType in [RESOURCEDISPLAYTYPE_DOMAIN, RESOURCEDISPLAYTYPE_NETWORK] then begin
               { Must recurse }
               if FlagIncluded(P^.dwUsage, RESOURCEUSAGE_CONTAINER) then begin
                  { Recursion possible }
                  Result := NetEnumerate(P, Scope, ResourceType, ComputerName);
                  if Result then break;
               end; { then }
            end; { then }
            //Move P to next entry in lpnrLocalc
            P := PNetResource(Integer(P) + SizeOf(TNetResource));
         end; { for }
      end; { NO_ERROR }

      ERROR_MORE_DATA : begin
         { The buffer is too small for even one entry:
           increase the buffer... }
         FreeMem(lpnrLocalc);
         lpnrLocalc := AllocMem(BufferSize);
         { ...and try again }
         end; { ERROR_MORE_DATA }

      else
      {
         ERROR_NO_MORE_ITEMS  (Enumeration is complete)
         ERROR_INVALID_HANDLE (The handle given by the hEnum parameter is not valid)
         ERROR_NO_NETWORK     (No network is present)
         ERROR_EXTENDED_ERROR (use WNetGetLastError for details)
      }
         break;
      end; { case }
   until Result;
   FreeMem(lpnrLocalc);
   WNetCloseEnum(hEnum);
end; { NetEnumerate }

{ #################### Disk Manipulation #################### }

function GetDiskFreeSpaceExA; external kernel32 name 'GetDiskFreeSpaceExA';
function GetDiskFreeSpaceExW; external kernel32 name 'GetDiskFreeSpaceExW';
function GetDiskFreeSpaceEx;  external kernel32 name 'GetDiskFreeSpaceExA';

{ #################### FileName Manipulation #################### }

function HasTrailer(const FileName : String) : Boolean;
{
   Returns True if FileName ends with a '\', False otherwise.
}
begin
   Result := (FileName <> '') and (FileName[Length(FileName)] = sBackslash);
end; { HasTrailer }

function AddTrailer(var FileName : String) : Boolean;
{
   AddTrailer adds the closing '\' (if missing) to FileName.
   Returns True if it was missing, False otherwise.
   An empty FileName is left unchanged (and returns False).
}
begin
   Result := not HasTrailer(FileName);
   if Result and (FileName <> '') then FileName := FileName + sBackslash;
end; { AddTrailer }

function RemoveTrailer(var FileName : String) : Boolean;
{
   RemoveTrailer removes the closing '\' (if any) from FileName.
   Returns True if there was one, False otherwise.
}
begin
   Result := HasTrailer(FileName);
   if Result then FileName := Copy(FileName, 1, Length(FileName) - 1);
end; { RemoveTrailer }

function IsDrive(const FileName : String) : Boolean;
{
   Returns True if FileName begins with <DriveLetter> + ':', False otherwise.
}
begin
   Result := (Length(FileName) >= 2) and (FileName[2] = ':') and (FileName[1] in DriveLetterSet);
end; { IsDrive }

function IsUNCServer(const FileName : String) : Boolean;
{
   Returns True if FileName begins with '\\', False otherwise.
}
begin
   Result := (Length(FileName) >= 2) and (FileName[1] = sBackslash) and (FileName[2] = sBackslash);
end; { IsUNCServer }

function ExtractFileTitle(const FileName : String) : String;
{
   ExtractFileTitle extracts the title part of the given
   filename.  The resulting string is the rightmost characters of FileName,
   starting with the first character after the colon or backslash that
   separates the path information from the name and stopping short of the
   period and extension.
   The resulting string is empty if FileName contains no title.
}
var
   i, j : Integer;
begin
   i := LastDelimiter('.', FileName);
   j := LastDelimiter('\:', FileName);
   if (i > j) then
      Result := Copy(FileName, j + 1, i - (j + 1))
   else if (j > 0) then
      Result := Copy(FileName, j + 1, MaxInt)
   else
      Result := FileName;
end; { ExtractFileTitle }

function ExtractUNCServer(const FileName: string): string;
{
   ExtractUNCServer extracts the UNC server part of the given filename.
   For filenames with a UNC path, the resulting string is in the form
   '\\<servername>'.
   Otherwise, the result is an empty string.
}
begin
   Result := '';
   if not IsUNCServer(FileName) then Exit;
   Result := Copy(FileName, 1, 1 + Pos(sBackslash, Copy(FileName + sBackslash, 3, Length(FileName))));
end; { ExtractUNCServer }

function ExtractUNCResource(const FileName: string): string;
{
   ExtractUNCResource extracts the UNC resource part of the given filename.
   For filenames with a UNC path, the resulting string is in the form
   '\<resourcename>'.
   Otherwise, the result is an empty string.
}
begin
   Result := ExtractUNCServer(FileName);
   if Result = '' then exit;
   //Remove UNC server part from FileName
   Result := Copy(FileName, Length(Result) + 1, Length(FileName));
   if (Result = '') or
      (Result = sBackslash)
   then begin
      Result := '';
   end else begin
      //Result is '\...'
      Result := Copy(Result, 1, Pos(sBackslash, Copy(Result + sBackslash, 2, Length(Result))));
   end; { else }
end; { ExtractUNCResource }

procedure ParseFileName(const FileName : String; var Drive : String; var Path : String; var Title : String; var Extension : String);
{
   Parses the FileName into its components (FileName need not exist).
   Drive will be either '', '<drive>:\' or '\\<server>\<resource>\'.
   Path will be either '', '<dir>\<dir>\<...>\' or a relative path.
   Title and Extension make up the file's name per se;
   Extension includes the period.  Either could be null strings.
   If FileName is a directory, behaviour depends on the closing '\'.

   Compare with Delphi's ProcessPath?
}
begin
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   Drive := ExtractFileDrive(FileName);
   AddTrailer(Drive);
   { Remove prefix Drive string }
   Path := Copy(Path, Length(Drive) + 1, Length(Path));
   Title := ExtractFileTitle(FileName);
   { ExtractFileExt returns either '' or '.<extension>' }
   Extension := ExtractFileExt(FileName);
end; { ParseFileName }

{ #################### File System Functions #################### }

function FileTimeToDateTime(FileTime : TFileTime; ToLocal : Boolean) : TDateTime;
{
   Converts the FileTime to a TDateTime value.
   When ToLocal is True, converts the FileTime from UTC to Local Time first.
   Returns -1 in case of failure.
}
var
// DOSDateTime : Integer;
   FileTime2   : TFileTime;
   SystemTime  : TSystemTime;
begin
   Result := -1;
   //Convert from UTC to local time?
   if ToLocal then begin
      if not FileTimeToLocalFileTime(FileTime, FileTime2) then Exit;
   end else begin
      FileTime2 := FileTime;
   end; { else }
(*
   //PATH A:
   //Convert to DOS date-time value and then to TDateTime
   if not FileTimeToDosDateTime(FileTime2, LongRec(DOSDateTime).Hi, LongRec(DOSDateTime).Lo) then Exit;
   try
      Result := FileDateToDateTime(DOSDateTime);
   except
      //Eat the exception
   end; { try }
*)
   //PATH B:
   //Convert to system time value and then to TDateTime
   if not FileTimeToSystemTime(FileTime2, SystemTime) then Exit;
   try
      Result := SystemTimeToDateTime(SystemTime);
   except
      //Eat the exception
   end; { try }
end; { FileTimeToDateTime }

function IsExtantFileContainer(const FileName : String) : Boolean;
{
   Returns True if FileName is a pure local drive ('<drive>:') or
   UNC drive ('\\<Server>[\<Resource>]') that exists.
   Inaccessible UNC servers and/or resources do not "exist".
   The function is not case-sensitive.
}
begin
   Result := False;
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   if (FileName <> ExtractFileDrive(FileName)) then Exit;
   if IsDrive(FileName) then begin
      Result := DriveExists(FileName);
   end else if ExtractUNCServer(FileName) = FileName then begin
      //Pure server
      Result := UNCServerExists(FileName);
   end else begin
      //Server+Resource
      Result := UNCResourceExists(FileName);
   end; { else }
end; { IsExtantFileContainer }

function GetLongFileName(FileName : String) : String;
{
   Obtains the Long version of a file name.
   Input must be fully qualified;
   Output will be.
   Returns an empty string if the file does not exist or is inaccessible.
}
var
   F       : TSearchRec;
   Trailer : String;
{
   You cannot use GetFullPathName because its purpose is to stick
   the current path in front of its argument, pure and simple.
}
begin
   Result := '';

   {
   Note: FindFirst fails if the FileName is a directory and the concluding backslash is included
   }
   if RemoveTrailer(FileName) then begin
      Trailer := sBackslash;
   end else begin
      Trailer := '';
   end; { else }
   if FileName = '' then Exit;

   //Must let pass drives, servers, resources
   //Note that server and share names are upper case in Win95 but not WinNT
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   if IsExtantFileContainer(FileName) then begin
      Result := FileName + Trailer;
      Exit;
   end; { then }

   //Does the file or directory exist?
   //An unsuccessfull FindFirst must not be FindClosed
   if (0 <> FindFirst(FileName, faAnyFile, F)) then Exit;

   //Count delimiters
   if (Pos(sBackslash, FileName) = LastDelimiter(sBackslash, FileName))
   then begin
      //File name is of "C:\FileName.Ext" type
      Result := Copy(FileName, 1, LastDelimiter(sBackslash, FileName)) + F.Name + Trailer;
   end else begin
      //File name is either of "C:\Path\FileName.ext" or "\\Server\Resource\Path\FileName.ext" type
      //Must recurse over path directories
      Result := GetLongFileName(Copy(FileName, 1, LastDelimiter(sBackslash, FileName) - 1)) + sBackslash + F.Name + Trailer;
   end; { else }
   FindClose(F);
end; { GetLongFileName }

function GetShortFileName(FileName : String) : String;
{
   Obtains the Short (8.3) version of a file name.
   Input must be fully qualified;
   Output will be.
   Returns an empty string if the file does not exist or is inaccessible.
}
var
   Trailer  : String;
   lpBuffer : packed array[0..MAX_PATH] of Char;
begin
   Result := '';

   if RemoveTrailer(FileName) then begin
      Trailer := sBackslash;
   end else begin
      Trailer := '';
   end; { else }
   if FileName = '' then Exit;

   //Must let pass drives, servers, resources
   //Note that server and share names are upper case in Win95 but not WinNT
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   if IsExtantFileContainer(FileName) then begin
      Result := ANSIUpperCase(FileName + Trailer);
      Exit;
   end; { then }

   {
   GetShortPathName will "localise" a UNC path whenever the UNC
   server is the local machine, just like GetLocalName does.
   It does NOT, however, localise imported paths (remote resources mapped to a local drive).
   }
   SetLastError(0);
   if 0 = GetShortPathName(PChar(FileName), lpBuffer, MAX_PATH) then Exit;
   Result := String(lpBuffer);
   //Keep the original drive information
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   if 0 <> ANSICompareText(ExtractFileDrive(FileName), ExtractFileDrive(Result)) then begin
      Result := ANSIUpperCase(ExtractFileDrive(FileName) + Copy(Result, 1 + Length(GetLocalName(ExtractFileDrive(FileName))), Length(Result)));
   end; { then }
   //GetShortPathName appends a trailing backslash when the input UNC name did not have one
   if Trailer = '' then RemoveTrailer(Result);
   Result := Result + Trailer;
end; { GetShortFileName }

function GetRemoteName(LocalName : String) : String;
{
   Returns the UNC path and title of the fully qualified LocalName.
   Returns an empty string if the local file isn't shared across
   the network or some other network-error occurs (such as password
   protection).
   GetRemoteName also works with remote files that have been mapped
   to a local drive letter.
   The returned path and title are in Long File Name format even
   if the input LocalName was in 8.3 format.
   Example:
   If your machine shares its C:\Temp directory as \\Avior\Temporary, then
   GetRemoteName('C:\Temp\Dummy.bak') returns '\\Avior\Temporary\Dummy.bak'.
   If you have mapped the remote resource '\\Pascal\Temporary\' to 'K:\', then
   GetRemoteName('K:\Dummy.bak') returns '\\Pascal\Temporary\Dummy.bak'.
}
var
   lpszRemoteName : PChar;
   i,
   BestMatch,
   lpcchBuffer    : Integer;
   WorkingName    : String;
   Resources,
   Paths          : TStrings;
begin
   //Make sure input is in long file name format
   Result := GetLongFileName(LocalName);
   if Result = '' then Exit;

   //Already an UNC name?
   if IsUNCServer(Result) then Exit;

   //Check for imported servers and shares first
{
   WNetGetConnection only works with local devices that are mapped remote devices.
   That is, it'll return the UNC of remote drives mapped to a local drive letter
   and the like, but won't give you the "exported" UNC of physically local devices.
   DUTGetSharedResourceNames supplies those.
}
   lpcchBuffer := 1024;
   GetMem(lpszRemoteName, lpcchBuffer);
   try { lpszRemoteName }
      //WNetGetConnection fails if LocalName goes any deeper than the
      //device (drive) level --We must pass only the drive letter and colon
      while True do begin
         SetLastError(0);
         if NO_ERROR = WNetGetConnection(PChar(Copy(Result, 1, 2)), lpszRemoteName, lpcchBuffer) then begin
            Result := String(lpszRemoteName) + Copy(Result, 3, Length(Result));
            break;
         end else begin
            case GetLastError of
            ERROR_MORE_DATA : begin
               //Buffer wasn't large enough; lpcchBuffer now contains required size
               FreeMem(lpszRemoteName);
               GetMem(lpszRemoteName, lpcchBuffer);
            end; { ERROR_MORE_DATA }
(*
            ERROR_BAD_DEVICE :
            ERROR_NOT_CONNECTED :
            ERROR_CONNECTION_UNAVAIL :
            ERROR_NO_NETWORK :
            ERROR_EXTENDED_ERROR :
            ERROR_NO_NET_OR_BAD_PATH :
*)
            else
               //Check for physically local shares
               Resources := TStringList.Create;
               try { Resources }
                  Paths := TStringList.Create;
                  try { Paths }
                     DUTGetSharedResourceNames(Resources, Paths);
                     //At this point, we have a list of share names; the
                     //corresponding Objects contain the local path strings
                     BestMatch := 0;
                     WorkingName := Result;
                     Result := '';
                     for i := 0 to (Resources.Count - 1) do begin
                        //Comparison should not be case-sensitive
                        //We trim the Paths[i] closing backslash in case the LocalName doesn't have one
                        if 0 = ANSICompareText(Copy(Paths[i], 1, Length(Paths[i]) - 1), Copy(WorkingName, 1, Length(Paths[i]) - 1)) then begin
                           //Since there may be several matches, we'll look for the one
                           //that has the longest Path (and thus very likely the shortest UNC name)
                           if Length(Paths[i]) > BestMatch then begin
                              BestMatch := Length(Paths[i]);
                              Result := Copy(Resources[i], 1, Length(Resources[i]) - 1) + Copy(WorkingName, Length(Paths[i]), Length(WorkingName));
                           end; { then }
                        end; { then }
                     end; { for }
                  finally
                     Paths.Free;
                  end; { try }
               finally
                  Resources.Free;
               end; { try }
               break;
            end; { case }
         end; { else }
      end; { while }
   finally
      FreeMem(lpszRemoteName);
   end; { try }
end; { GetRemoteName }

function GetLocalName(UNCName : String) : String;
{
   Returns the local path and title of the UNCName.
   Returns an empty string if the file isn't truly local or if
   the remote resource has not been mapped to a local file.
   The returned path and title are in Long File Name format even
   if the input UNCName was in 8.3 format.
   Example:
   If your machine shares its C:\Temp directory as \\Avior\Temporary, then
   GetLocalName('\\Avior\Temporary\Dummy.bak') returns 'C:\Temp\Dummy.bak'.
   If you have mapped the remote resource '\\Pascal\Temporary\' to 'K:\', then
   GetLocalName('\\Pascal\Temporary\Dummy.bak') returns 'K:\Dummy.bak'.
}
var
   Letter       : TDriveLetter;
// lpBufferSize : DWORD;
// lpBuffer     : Pointer;
   UNC          : String;
   Resources,
   Paths        : TStrings;
   i            : Integer;
begin
   Result := GetLongFileName(UNCName);
   if Result = '' then Exit;

   //Already a local name?
   if IsDrive(Result) then Exit;

   //Pure server name?
   if (Result = ExtractUNCServer(Result)) or
      (Result = (ExtractUNCServer(Result) + sBackslash))
   then begin
      Result := '';
      Exit;
   end; { then }

   //Is there a local mapping for the UNC resource/share?
   for Letter := High(TDriveLetter) downto Low(TDriveLetter) do begin
      if DRIVE_REMOTE = GetDriveType(PChar(Letter + ':\')) then begin
         //We would normally use WNetGetUniversalName to find out the UNC
         //name the drive (Letter + ':\') is mapped from, but there is a
         //bug in the Win95 version that makes it unusable.
(*
         lpBufferSize := 512;
         GetMem(lpBuffer, lpBufferSize);
         try { lpBuffer }
            while True do begin
               SetLastError(0);
               if NO_ERROR = WNetGetUniversalName(PChar(String(Letter + ':')), UNIVERSAL_NAME_INFO_LEVEL, lpBuffer, lpBufferSize) then begin
                  UNC := PUniversalNameInfo(lpBuffer).lpUniversalName;
                  if 0 = ANSICompareText(UNC, Copy(Result, 1, Length(UNC))) then begin
                     //Found it!
                     Result := String(Letter + ':') + Copy(Result, Length(UNC), Length(Result));
                     Exit;
                  end else begin
                     //Not this drive: Just fail
                     break;
                  end; { else }
               end else begin
                  case GetLastError of
                  ERROR_MORE_DATA : begin
                     //Buffer wasn't large enough; lpBufferSize now contains required size
                     FreeMem(lpBuffer);
                     GetMem(lpBuffer, lpBufferSize);
                  end; { ERROR_MORE_DATA }
                  else
                     //Just fail --maybe raise an exception?
                     break;
                  end; { case }
               end; { else }
            end; { while }
         finally
            FreeMem(lpBuffer);
         end; { try }
*)
         UNC := ExpandUNCFileName(Letter + ':\');
         //Compare common length
         if (UNC <> '') and
            (0 = ANSICompareText(Copy(UNC, 1, Length(Result)), Copy(Result, 1, Length(UNC))))
         then begin
            //Found it!
            Result := String(Letter + ':\') + Copy(Result, 1 + Length(UNC), Length(Result));
            //Remove trailing backslash if required
            if not HasTrailer(UNCName) then RemoveTrailer(Result);
            Exit;
         end; { then }
      end; { then }
   end; { for }

   //Now check exported local folders
   {
   In turns out GetShortPathName will "localise" a UNC path whenever
   the UNC server is the local machine, so we could use it here.
   However, this ain't guaranteed, so we take no chances.
   }
   Resources := TStringList.Create;
   try { Resources }
      Paths := TStringList.Create;
      try { Paths }
         DUTGetSharedResourceNames(Resources, Paths);
         for i := 0 to (Resources.Count - 1) do begin
            //Comparison should not be case-sensitive
            //We trim the Resources[i] closing backslash in case the UNCName doesn't have one
            if 0 = ANSICompareText(Copy(Resources[i], 1, Length(Resources[i]) - 1), Copy(Result, 1, Length(Resources[i]) - 1)) then begin
               Result := Copy(Paths[i], 1, Length(Paths[i]) - 1) + Copy(Result, Length(Resources[i]), Length(Result));
               Exit;
            end; { then }
         end; { for }
      finally
         Paths.Free;
      end; { try }
   finally
      Resources.Free;
   end; { try }

   //No mapping succeeded
   Result := '';
end; { GetLocalName }

function DirExists(const Dir : String) : Boolean;
{
   Determines whether the Dir exists.
   Dir can be a relative path.
   If Dir is a *file* or an empty string, returns False.
   Works equally well with and without the closing '\'.
   The function is not case-sensitive.
}
var
   Tmp   : String;
begin
   Result := False;
   if Dir = '' then exit;
   { Note current directory }
   GetDir(0, Tmp);
   { Attempt to change to it }
   Result := SetCurrentDirectory(PChar(Dir));
   { Restore current directory }
   SetCurrentDirectory(PChar(Tmp));
end; { DirExists }

function DriveExists(const Drive : String) : Boolean;
{
   Returns True if the Drive exists, False otherwise.
   The function is not case-sensitive.
}
begin
   Result := not (GetDriveType(PChar(Drive[1] + ':\')) in [DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR]);
end; { DriveExists }

function UNCResourceExists(const ResourceName : String) : Boolean;
{
   Returns True if the ResourceName exists on the network.
   ResourceName should be in the form "\\Server\Resource"
   (any appended path will be ignored).
   The function is not case-sensitive.
}
var
   F : TSearchRec;
begin
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   SetLastError(0);
   //An unsuccessfull FindFirst must not be FindClosed
   case FindFirst(ExtractFileDrive(ResourceName) + '\*.*', faAnyFile, F) of
   0 : begin
      //Success!
      FindClose(F);
      Result := True;
      end; { 0 }
(*
   ERROR_BAD_DEV_TYPE =  66, //Argument is Server+Resource only; existence of either undefined
   ERROR_BAD_NET_NAME =  67, //Server exists, but the Resource does not
   ERROR_BAD_NETPATH  =  53, //Server does not exist (resource was specified)
   ERROR_BAD_PATHNAME = 161, //No Resource in argument; Server existence undefined
*)
   else
      Result := False;
{$IFDEF Debug}
MessageDlg(Format('Error %d: <<%s>>', [GetLastError, SysErrorMessage(GetLastError)]), mtCustom, [mbOK], 0);
{$ENDIF}
   end; { case }
end; { UNCResourceExists }

function UNCServerExists(const ServerName : String) : Boolean;
{
   Returns True if the ServerName exists, False otherwise.
   ServerName should be of the form "\\Name"
   (an appended resource will be used to speed up the network request,
   any appended path will be ignored).
   The function is not case-sensitive.
   There may be a substantial delay if ServerName does not exist (or is inaccessible).
}
var
   F : TSearchRec;
   S : String;
begin
   //Is ServerName self?
   Result := (0 = ANSICompareText('\\' + DUTGetComputerServerName, ExtractUNCServer(ServerName)));
   if Result then Exit;

   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   S := ExtractFileDrive(ServerName);
   if S = ExtractUNCServer(S) then S := S + '\ABC';
   SetLastError(0);
   //An unsuccessfull FindFirst must not be FindClosed
   case FindFirst(S, faAnyFile, F) of
   0 : begin
      //Should not happen
      FindClose(F);
      Result := True;
      end; { 0 }
// 2106,                      //Server is self
   ERROR_BAD_DEV_TYPE,        //Argument is Server+Resource only; existence of either undefined
   ERROR_BAD_NET_NAME : begin //Server exists, but the Resource does not
      //We'll get a ERROR_BAD_DEV_TYPE if \\<ServerName>\ABC exists
      //We'd get a ERROR_BAD_NETPATH instead if the Server did not exist
      Result := True;
      end; { ERROR_BAD_PATHNAME }
(*
   ERROR_BAD_DEV_TYPE =  66, //Argument is Server+Resource only; existence of either undefined
   ERROR_BAD_NET_NAME =  67, //Server exists, but the Resource does not
   ERROR_BAD_NETPATH  =  53, //Server does not exist (resource was specified)
   ERROR_BAD_PATHNAME = 161, //No Resource in argument; Server existence undefined
*)
   else
      //Expect ERROR_BAD_NETPATH otherwise
      Result := False;
{$IFDEF Debug}
MessageDlg(Format('Error %d: <<%s>>', [GetLastError, SysErrorMessage(GetLastError)]), mtCustom, [mbOK], 0);
{$ENDIF}
   end; { case }
end; { UNCServerExists }

function UNCServerConnected(const ServerName : String; Live : Boolean) : Boolean;
{
   Returns True if the UNC server ServerName is currently connected through the network.
   When Live is True, only live connections are checked; otherwise all persistent
   connections are checked.
   ServerName must be of the form "\\Name".
   The function is not case-sensitive.
}
begin
   { Start enumeration at the root of the network }
   if Live then begin
      Result := NetEnumerate(Nil, RESOURCE_CONNECTED, RESOURCETYPE_DISK, ServerName);
   end else begin
      Result := NetEnumerate(Nil, RESOURCE_REMEMBERED, RESOURCETYPE_DISK, ServerName);
   end; { else }
end; { UNCServerConnected }

function RemoteComputerExists(const ComputerName : String) : Boolean;
{
   Returns True if the remote computer ComputerName exists on the network.
   ComputerName must be of the form "\\Name".
   The function is not case-sensitive.
   Be warned that this may take a *long* time to return.
   (Much longer than UNCServerExists)

   Based on code supplied by:
   Michael P. Bobowski
   Atled Engineering Group
   Milwaukee, WI
   <mbobo@execpc.com>
}
begin
   { Start enumeration at the root of the network }
   Result := NetEnumerate(Nil, RESOURCE_GLOBALNET, RESOURCETYPE_DISK, ComputerName);
end; { RemoteComputerExists }
(*
function GetRemoteComputers : TStringList;
{
   Returns a TStringList (possibly empty) giving the remote computer names on the network.
   The names are all of the form "\\Name".  The TStringList is not sorted.
   Be warned that this may take a *long* time to return.
}
   procedure NetEnumerate(lpnr : PNetResource; Scope, ResourceType : DWORD; var ComputerNames : TStringList);
   var
      hEnum        : THandle;
      BufferSize,
      NumEntries,
      Entry        : Integer;
      lpnrLocalc,  //Actually a pointer to an Array of TNetResource
      P            : PNetResource;
   begin
      if NO_ERROR <> WNetOpenEnum(Scope, ResourceType, 0 { Usage: All resources }, lpnr, hEnum) then begin
         Exit;
      end;
      BufferSize := 16384; { 16 kB }
      lpnrLocalc := AllocMem(BufferSize);
      repeat
         { Get as many entries as possible; NumEntries will be set to the number actually read (if successfull) }
         NumEntries := $FFFFFFFF;
         case WNetEnumResource(hEnum, NumEntries, lpnrLocalc, BufferSize) of
         NO_ERROR : begin
            P := lpnrLocalc;
            for Entry := 0 to (NumEntries - 1) do begin
               { P^.dwScope will be Scope since that is what we asked for }
               if P^.dwDisplayType = RESOURCEDISPLAYTYPE_SERVER then begin
                  { RESOURCEUSAGE_CONTAINER = (P^.dwUsage and RESOURCEUSAGE_CONTAINER) should also be True }
                  ComputerNames.Add(P^.lpRemoteName);
               end else if P^.dwDisplayType in [RESOURCEDISPLAYTYPE_DOMAIN, RESOURCEDISPLAYTYPE_NETWORK] then begin
                  { Must recurse }
                  if RESOURCEUSAGE_CONTAINER = (P^.dwUsage and RESOURCEUSAGE_CONTAINER) then begin
                     { Recursion possible }
                     NetEnumerate(P, Scope, ResourceType, ComputerNames);
                  end; { then }
               end; { then }
               //Move P to next entry in lpnrLocalc
               P := PNetResource(Integer(P) + SizeOf(TNetResource));
            end; { for }
         end; { NO_ERROR }

         ERROR_MORE_DATA : begin
            { The buffer is too small for even one entry:
              increase the buffer... }
            FreeMem(lpnrLocalc);
            lpnrLocalc := AllocMem(BufferSize);
            { ...and try again }
            end; { ERROR_MORE_DATA }

         else
         {
         ERROR_NO_MORE_ITEMS  (Enumeration is complete)
         ERROR_INVALID_HANDLE (The handle given by the hEnum parameter is not valid)
         ERROR_NO_NETWORK     (No network is present)
         ERROR_EXTENDED_ERROR (use WNetGetLastError for details)
         }
            break;
         end; { case }
      until False;
      FreeMem(lpnrLocalc);
      WNetCloseEnum(hEnum);
   end; { NetEnumerate }

begin
   Result := TStringList.Create;
   { Start enumeration at the root of the network }
   NetEnumerate(Nil, RESOURCE_GLOBALNET, RESOURCETYPE_DISK, Result);
end; { GetRemoteComputers }
*)
function IsALink(FileName : String) : Boolean;
{
   Returns True if the FileName is an LNK file.
   The function is not case-sensitive.
}
begin
   Result := (Length(FileName) > 3) and
      (Uppercase(Copy(FileName, Length(FileName) - 3, 4)) = '.LNK');
end; { IsALink }

function DUTGetComputerName : String;
{
   Obtains the local machine's name.
}
var
   Size : Integer;
   Name : String;
begin
   Result := '';
   Size := 1 + MAX_COMPUTERNAME_LENGTH;
   Name := StringOfChar(#0, Size);
   if not GetComputerName(PChar(Name), Size) then Exit;
   Result := Copy(Name, 1, Size);
end; { DUTGetComputerName }

function DUTGetComputerServerName : String;
{
   Obtains the local machine's network server name.
   The result does not include the leading '\\'.
}
var
   R : TRegistry;
begin
{
   The local machine's server name is obtained by looking up the 
   REGSTR_PATH_VNETSUP registry key:
   HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\VxD\VNETSUP
   and reading the REGSTR_VAL_COMPUTERNAME ('ComputerName') string value.
}
   Result := '';
   R := TRegistry.Create;
   try { R }
      R.RootKey := HKEY_LOCAL_MACHINE;
      if R.OpenKey(sBackslash + REGSTR_PATH_VNETSUP, False) then begin
         Result := R.ReadString(REGSTR_VAL_COMPUTERNAME);
         R.CloseKey;
      end; { then }
   finally
      R.Free;
   end; { try }
end; { DUTGetComputerServerName }

procedure DUTGetSharedResourceNames(var Resources, Paths : TStrings);
{
   Resources is Cleared and then loaded with the names
   of the UNC resources shared by the local computer,
   prepended with the local computer's UNC server name.
   Resources is also Sorted.
   Paths is likewise Cleared and then loaded with the
   local paths of each resource.
   Both sets of strings include the closing backslashes.
   IMPORTANT: Make sure both TStrings' Sorted property are False.
   Example:
   Resources[0] == '\\Avior\Temporary\'    Paths[0] == 'C:\Temp\'
}
var
   R    : TRegistry;
   i    : Integer;
   Path : String;
begin
   Resources.Clear;
   Paths.Clear;
{
   To recognise physically local devices that have been "exported", you must check
   the REGSTR_KEY_SHARES Registry key:
   HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Network\LanMan
   It'll have a subkey for each exported device.  The subkey's name is the
   device's UNC name; the subkey's REGSTR_VAL_SHARES_PATH ('Path') string value
   is the device's corresponding local path.
   For example, the subkey "AVIOR_C" may have a Path = "C:\".
}
   R := TRegistry.Create;
   try { R }
      R.RootKey := HKEY_LOCAL_MACHINE;
      if R.OpenKey(sBackslash + REGSTR_KEY_SHARES, False) then begin
         R.GetKeyNames(Resources);
         (Resources as TStringList).Sort;
         i := 0;
         while i < Resources.Count do begin
            if R.OpenKey(sBackslash + REGSTR_KEY_SHARES + sBackslash + Resources[i], False)
            then begin
               //Make sure the Path is in long file name format
               Path := GetLongFileName(R.ReadString(REGSTR_VAL_SHARES_PATH));
               //Synchronise trailing backslashes
               //(Key names can't have any)
               AddTrailer(Path);
               Paths.Add(Path);
               Resources[i] := '\\' + DUTGetComputerServerName + sBackslash + Resources[i] + sBackslash;
               Inc(i);
            end else begin
               Resources.Delete(i);
            end; { else }
         end; { while }
         R.CloseKey;
      end; { then}
   finally
      R.Free;
   end; { try }
end; { DUTGetSharedResourceNames }

function DUTGetSecurityProvider : String;
{
   Obtains the local machine's network security provider container name.
   This may be different from the name of the domain (workgroup) the
   local machine belongs to (see DUTGetUserDomain).
}
var
   R : TRegistry;
begin
{
   For Win95,
   The local machine's network security provider container name is
   obtained by looking up the registry key:
   HKEY_LOCAL_MACHINE\Security\Provider
   and its 'Container' string value.
   It is also under the key:
   HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\MSNP32\NetworkProvider
   and its AuthenticatingAgent string value.

   Neither key works for WinNT.
}
   Result := '';
   R := TRegistry.Create;
   try { R }
      R.RootKey := HKEY_LOCAL_MACHINE;
      if R.OpenKey('\Security\Provider', False) then begin
         Result := R.ReadString('Container');
         R.CloseKey;
      end; { then }
   finally
      R.Free;
   end; { try }
end; { DUTGetSecurityProvider }

function DUTGetUserDomain : String;
{
   Obtains the domain the computer belongs to.
   This is slow!
}
   function GUDNetEnumerate(lpnr : PNetResource; Scope, ResourceType : DWORD; const ComputerName : String; var DomainName : String) : Boolean;
   {
      Private network enumeration recursive function.
      lpnr should initially be Nil to start the enumeration at the network's root.
      Scope should be RESOURCE_GLOBALNET, RESOURCE_CONNECTED or RESOURCE_REMEMBERED.
      ResourceType should be RESOURCETYPE_ANY, RESOURCETYPE_DISK or RESOURCETYPE_PRINT.
   }
   var
      hEnum        : THandle;
      BufferSize,
      NumEntries,
      Entry        : Integer;
      lpnrLocalc,  //Actually a pointer to an Array of TNetResource
      P            : PNetResource;
   begin
      Result := False;
      if NO_ERROR <> WNetOpenEnum(Scope, ResourceType, RESOURCEUSAGE_CONTAINER, lpnr, hEnum) then Exit;
      try { hEnum }
         BufferSize := 16384; { 16 Kib }
         lpnrLocalc := AllocMem(BufferSize);
         try { lpnrLocalc }
            repeat
               { Get as many entries as possible; NumEntries will be set to the number actually read (if successfull) }
               NumEntries := $FFFFFFFF;
               case WNetEnumResource(hEnum, NumEntries, lpnrLocalc, BufferSize) of
               NO_ERROR : begin
                  P := lpnrLocalc;
                  for Entry := 0 to (NumEntries - 1) do begin
                     { P^.dwScope will be Scope since that is what we asked for }
                     {
                     Compare the lpRemoteName up to ComputerName's length;
                     since lpRemoteName always includes a Resource part,
                     we must do this to correctly recognise Servers
                     }
                     if (P^.dwDisplayType = RESOURCEDISPLAYTYPE_SERVER) and
                        (0 = ANSICompareText(Copy(P^.lpRemoteName, 1, Length(ComputerName)), ComputerName))
                     then begin
                        { FlagIncluded(P^.dwUsage, RESOURCEUSAGE_CONTAINER) should also be True }
                        Result := True;
                        break;
                     end; { then }
                     { ResourceType is irrelevant }

                     if P^.dwDisplayType in [RESOURCEDISPLAYTYPE_DOMAIN, RESOURCEDISPLAYTYPE_NETWORK] then begin
                        { Must recurse }
//                      if FlagIncluded(P^.dwUsage, RESOURCEUSAGE_CONTAINER) then begin
                           { Recursion possible }
                           Result := GUDNetEnumerate(P, Scope, ResourceType, ComputerName, DomainName);
                           if Result then begin
                              {
                              When recursing down from a container, if we found
                              the ComputerName, then the container's RemoteName
                              is what we were looking for.
                              If DomainName is already non-blank, then we're at
                              the level of one of the container's containers.
                              }
                              if DomainName = '' then DomainName := P^.lpRemoteName;
                              break;
                           end; { then }
//                      end; { then }
                     end; { then }
                     //Move P to next entry in lpnrLocalc
                     P := PNetResource(Integer(P) + SizeOf(TNetResource));
                  end; { for }
               end; { NO_ERROR }

               ERROR_MORE_DATA : begin
                  { The buffer is too small for even one entry:
                    increase the buffer... }
                  FreeMem(lpnrLocalc);
                  BufferSize := 2*BufferSize;
                  lpnrLocalc := AllocMem(BufferSize);
                  { ...and try again }
               end; { ERROR_MORE_DATA }

               else
               {
                  ERROR_NO_MORE_ITEMS  (Enumeration is complete)
                  ERROR_INVALID_HANDLE (The handle given by the hEnum parameter is not valid)
                  ERROR_NO_NETWORK     (No network is present)
                  ERROR_EXTENDED_ERROR (use WNetGetLastError for details)
               }
                  break;
               end; { case }
            until Result;
         finally
            FreeMem(lpnrLocalc);
         end; { try }
      finally
         WNetCloseEnum(hEnum);
      end; { try }
   end; { GUDNetEnumerate }
begin
   //We must use RESOURCE_GLOBALNET otherwise the enumeration will
   //restrict itself to the local computer's mapped drives only.
   GUDNetEnumerate(Nil, RESOURCE_GLOBALNET, RESOURCETYPE_ANY, '\\' + DUTGetComputerName, Result);
end; { DUTGetUserDomain }

function DUTGetFileSize(const FileName: string) : Integer;
{
   Returns the FileName's size, in bytes.
   -1 if it does not exist (or is inaccessible), 0 if empty or a directory.
   Huge files (whose size exceeds $7FFF FFFF bytes)
   will raise an EIntOverflow exception.
}
var
   Handle      : THandle;
   FindData    : TWin32FindData;
   WorkingName : String;
begin
   Result := -1; //File does not exist
   WorkingName := FileName;
   RemoveTrailer(WorkingName);
   if WorkingName = '' then Exit;
   Handle := Windows.FindFirstFile(PChar(WorkingName), FindData);
   if Handle = INVALID_HANDLE_VALUE then Exit;
   Windows.FindClose(Handle);
   Result := 0; //File is a directory
   if FlagIncluded(FindData.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY) then Exit;
// True Size == (FindData.nFileSizeHigh * MAXDWORD { $FFFFFFFF } ) + FindData.nFileSizeLow;
   //Delphi LongInt is limited to a maximum of $7FFFFFFF
   Result := FindData.nFileSizeLow;
   if Result < 0 then raise EIntOverflow.Create(SIntOverflow);
end; { DUTGetFileSize }

procedure DUTGetDiskFreeSpace(const DriveName: string; var SectorsPerCluster, BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters : Integer);
{
   Returns the DriveName's vital space statistics.
   DriveName must be either a drive (X:\) or an UNC resource name
   (\\Server\Resource; pure UNC servers will fail).
   Normally, UNC paths would fail, but the function circumvents this
   by temporarily mapping the remote resource to a local drive letter.
   The function is not case-sensitive.
}
var
   MappedDrive,
   Drive        : String;
   Letter       : TDriveLetter;
begin
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   Drive := ExtractFileDrive(DriveName) + sBackslash;
   if (Drive = sBackslash) or (Drive = '\\\') then raise Exception.CreateFmt('"%s" is not a drive name.', [DriveName]);

   MappedDrive := '';
   //UNC path?
   if IsUNCServer(Drive) then begin
      //Pure server name?
      if (ExtractUNCServer(Drive) + sBackslash) = Drive then raise Exception.CreateFmt('The UNC path "%s" is missing its resource part.', [DriveName]);

      //Has the resource been mapped to a local drive?
      if MapDrive(Drive, True, Letter) then begin
         Drive := Letter + ':\';
      end else begin
         MappedDrive := Drive;
         Letter := High(TDriveLetter);
         if MapDrive(Drive, False, Letter) then begin
            Drive := Letter + ':\';
         end else begin
            MappedDrive := '';
         end; { else }
      end; { else }
   end; { then }
   try
      SetLastError(0);
      if not GetDiskFreeSpace(PChar(Drive), SectorsPerCluster, BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters) then
         raise Exception.CreateFmt('GetDiskFreeSpace failed.  Error #%d: %s', [GetLastError, SysErrorMessage(GetLastError)]);
   finally
      //Cleanup: Must un-map the MappedDrive
      if MappedDrive <> '' then begin
         //Ignore failure
         WNetCancelConnection(PChar(Copy(Drive, 1, 2)), True);
         //By disconnecting the drive (X:) instead of the resource (\\XYZ)
         //we do not break any eventual previous mappings of the resource
      end; { then }
   end; { try }
end; { DUTGetDiskFreeSpace }

function GetFileFootprint(const FileName: string) : Integer;
{
   Returns the FileName's footprint, in bytes.
   FileName must be a fully qualified path and title.
   The footprint will be a whole multiple of the underlying drive's cluster size.
   -1 if the file does not exist or is inaccessible.
   Directories and drives (local or remote) report a zero footprint.
   The function is not case-sensitive.
}
var
   BytesPerCluster,
   SectorsPerCluster,
   BytesPerSector,
   NumberOfFreeClusters,
   TotalNumberOfClusters : DWORD;
begin
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   DUTGetDiskFreeSpace(ExtractFileDrive(FileName), SectorsPerCluster, BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters);
   BytesPerCluster := SectorsPerCluster*BytesPerSector;
   {
   0 bytes            = 0 clusters occupied
   1..BytesPerCluster = 1 cluster occupied
   and so on...
   }
   Result := BytesPerCluster*((DUTGetFileSize(FileName) + (BytesPerCluster - 1)) div BytesPerCluster);
end; { GetFileFootprint }

function CanFileFit(const FileName, Drive : string) : Boolean;
{
   Whether the FileName can fit on Drive.
   The decision is based on the footprint, not the byte-size.
   FileName must be a fully qualified path and title.
   Drive may be a local drive ('X:\') or a UNC server and resource
   (it'll be temporarily mapped to a drive letter if need be).
   Returns False if the file does not exist or is inaccessible.
   The function is not case-sensitive.
}
var
   BytesPerCluster,
   SectorsPerCluster,
   BytesPerSector,
   NumberOfFreeClusters,
   TotalNumberOfClusters : DWORD;
begin
   DUTGetDiskFreeSpace(Drive, SectorsPerCluster, BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters);
   BytesPerCluster := SectorsPerCluster*BytesPerSector;
   {
   0 bytes            = 0 clusters occupied
   1..BytesPerCluster = 1 cluster occupied
   and so on...
   }
   Result := ((DUTGetFileSize(FileName) + (BytesPerCluster - 1)) div BytesPerCluster) <= NumberOfFreeClusters;
end; { CanFileFit }

function MapDrive(const DriveName : String; ExtantOnly : Boolean; var Letter : TDriveLetter) : Boolean;
{
   Sets Letter to the DriveName's mapped drive letter.
   DriveName must be an UNC resource name (\\Server\Resource; pure UNC servers will fail).
   When ExtantOnly is True, MapDrive sets Letter to the pre-existing mapping.
   If there is none, MapDrive fails.
   When ExtantOnly is False, MapDrive sets Letter to the pre-existing mapping.
   If there is none, MapDrive grabs the Letter passed in if it is free.
   If that also fails (the Letter is already taken), it works its way back up the
   alphabet (from Z towards A) until it finds a free Letter or runs out of letters.
   MapDrive returns True if successfull, False otherwise.
   The function is not case-sensitive.
}
var
   MappedDrive,
   Drive  : String;
begin
   Result := False;
   { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
   Drive := ExtractFileDrive(DriveName) + sBackslash;
   if (Drive = sBackslash) or
      (Drive = '\\\') or
      //UNC path?
      (not IsUNCServer(Drive)) or
      //Pure server name (No resource part)?
      ((ExtractUNCServer(Drive) + sBackslash) = Drive)
   then
      Exit;

   //Has the resource been mapped to a local drive?
   Drive := GetLocalName(Drive);
   if Drive <> '' then begin
      Letter := Drive[1];
      Result := True;
      Exit;
   end else begin
      if ExtantOnly then Exit;
      //Must map
      //Look for available Letter, starting with value passed in
      while DRIVE_NO_ROOT_DIR <> GetDriveType(PChar(Letter + ':\')) do begin
         if Letter > Low(TDriveLetter) then begin
            Dec(Letter);
         end else begin
            //Ran out of letters --failure
            Exit;
         end; { else }
      end; { then }
      Drive := Letter + ':\';
      //Found a drive letter to use
      { ExtractFileDrive returns either '', '\\', '<drive>:' or '\\<servername>[\<sharename>]' }
      MappedDrive := ExtractFileDrive(DriveName) + sBackslash;
      SetLastError(0);
      if NO_ERROR <> WNetAddConnection(PChar(MappedDrive), Nil, PChar(Copy(Drive, 1, 2))) then begin
         raise Exception.CreateFmt('WNetAddConnection failed.  Error #%d: %s', [GetLastError, SysErrorMessage(GetLastError)]);
      end; { then }
   end; { else }
end; { MapDrive }

function DUTGetWindowsDirectory : String;
{
   Obtains the Windows directory path.
   The ending backslash is included.
}
var
   Size : DWORD;
begin
   Size := MAX_PATH;
   Result := StringOfChar(#0, Size);
   Size := Windows.GetWindowsDirectory(PChar(Result), Size);
   if Size <> 0 then begin
      SetLength(Result, Size);
      AddTrailer(Result);
   end else begin
      Result := '';
   end; { else }
end; { DUTGetWindowsDirectory }

function DUTGetSystemDirectory : String;
{
   Obtains the Windows System directory path.
   The ending backslash is included.
}
var
   Size : DWORD;
begin
   Size := MAX_PATH;
   Result := StringOfChar(#0, Size);
   Size := Windows.GetSystemDirectory(PChar(Result), Size);
   if Size <> 0 then begin
      SetLength(Result, Size);
      AddTrailer(Result);
   end else begin
      Result := '';
   end; { else }
end; { DUTGetSystemDirectory }

function DUTGetCurrentDirectory : String;
{
   Obtains the current directory path.
   The ending backslash is included.
}
begin
   Result := GetCurrentDir;
   AddTrailer(Result);
end; { DUTGetCurrentDirectory }

function DUTGetTempDirectory : String;
{
   Obtains the Temporary directory path.
   The ending backslash is included.
}
var
   Size : DWORD;
begin
   Size := MAX_PATH;
   Result := StringOfChar(#0, Size);
   //Notice how GetTempPath has reversed the sequence of parameters compared to GetWindowsDirectory et al.
   Size := Windows.GetTempPath(Size, PChar(Result));
   if Size <> 0 then begin
      SetLength(Result, Size);
      AddTrailer(Result);
   end else begin
      Result := '';
   end; { else }
end; { DUTGetTempDirectory }

function DUTGetTempFilename(Path, Prefix : String) : String;
{
   Obtains a Temporary filename.
   The file exists upon return; it is created in Path (or the
   current temporary directory, if not specified).
   If Prefix is non-blank, the first three ANSI characters
   are used to begin the file title.
   The fully qualified path of the temporary file is returned.
}
begin
   Result := StringOfChar(#0, MAX_PATH);
   if Path = '' then Path := DUTGetTempDirectory;
   SetLastError(0);
   if 0 = GetTempFileName(PChar(Path), PChar(Prefix), 0, PChar(Result)) then
      raise Exception.CreateFmt('GetTempFileName failed.  Error #%d: %s', [GetLastError, SysErrorMessage(GetLastError)]);
   SetLength(Result, Pos(#0, Result) - 1);
end; { DUTGetTempFilename }

function GetSpecialFolderPath(SpecialFolderType : TSpecialFolderType) : String;
{
   One could also go via the registry, which is what
   DUTShellLink's GetSpecialFolderLocation does.
}
const
   sfToCSIDL_WinNT : Array[TSpecialFolderType] of Integer = (
      CSIDL_APPDATA,            // <user name>\Application Data
      CSIDL_COMMON_APPDATA,     // All Users\Application Data
      CSIDL_COMMON_DESKTOPDIRECTORY, // All Users\Desktop
      CSIDL_COMMON_PROGRAMS,    // All Users\Programs
      CSIDL_COMMON_STARTMENU,   // All Users\Start Menu
      CSIDL_COMMON_STARTUP,     // All Users\Startup
      CSIDL_COMMON_FAVORITES,   // All Users\Favorites
      CSIDL_DESKTOPDIRECTORY,   // <user name>\Desktop
      CSIDL_FAVORITES,          // <user name>\Favorites
      CSIDL_FONTS,              // Windows\Fonts
      CSIDL_NETHOOD,            // <user name>\NetHood
      CSIDL_PERSONAL,           // My Documents
      CSIDL_PRINTHOOD,          // Windows\PrintHood
      CSIDL_PROGRAMS,           // Start Menu\Programs
      CSIDL_RECENT,             // <user name>\Recent
      CSIDL_SENDTO,             // <user name>\SendTo
      CSIDL_STARTMENU,          // <user name>\Start Menu
      CSIDL_STARTUP,            // Start Menu\Programs\Startup
      CSIDL_TEMPLATES,          // Windows\ShellNew
      CSIDL_INTERNET_CACHE,     // Windows\Temporary Internet Files
      CSIDL_COOKIES,            // Windows\Profiles\<username>\Cookies
      CSIDL_HISTORY,            // Windows\Profiles\<username>\History
//    CSIDL_PROGRAM_FILES,      // Program Files
//    CSIDL_MYPICTURES,         // My Pictures
      CSIDL_WINDOWS,            // Windows
      CSIDL_SYSTEM              // Windows\System
//    CSIDL_LOCAL_APPDATA,      // <user name>\Local Settings\Application Data (non roaming)
//    CSIDL_ALTSTARTUP,         // non localized startup
//    CSIDL_COMMON_ALTSTARTUP,  // non localized common startup
//    CSIDL_PROFILE            // UserProfile
   );

   sfToCSIDL_Win95 : Array[TSpecialFolderType] of Integer = (
      CSIDL_APPDATA,            // <user name>\Application Data
      CSIDL_APPDATA,            // <user name>\Application Data
      CSIDL_DESKTOP,            // Windows\Desktop
      CSIDL_PROGRAMS,           // Start Menu\Programs
      CSIDL_STARTMENU,          // <user name>\Start Menu
      CSIDL_STARTUP,            // Start Menu\Programs\Startup
      CSIDL_FAVORITES,          // Windows\Favorites
      CSIDL_DESKTOPDIRECTORY,   // <user name>\Desktop
      CSIDL_FAVORITES,          // <user name>\Favorites
      CSIDL_FONTS,              // Windows\Fonts
      CSIDL_NETHOOD,            // <user name>\NetHood
      CSIDL_PERSONAL,           // My Documents
      CSIDL_PRINTHOOD,          // Windows\PrintHood
      CSIDL_PROGRAMS,           // Start Menu\Programs
      CSIDL_RECENT,             // <user name>\Recent
      CSIDL_SENDTO,             // <user name>\SendTo
      CSIDL_STARTMENU,          // <user name>\Start Menu
      CSIDL_STARTUP,            // Start Menu\Programs\Startup
      CSIDL_TEMPLATES,          // Windows\ShellNew
      CSIDL_INTERNET_CACHE,     // Windows\Temporary Internet Files
      CSIDL_COOKIES,            // Windows\Profiles\<username>\Cookies
      CSIDL_HISTORY,            // Windows\Profiles\<username>\History
//    CSIDL_PROGRAM_FILES,      // Program Files
//    CSIDL_MYPICTURES,         // My Pictures
      CSIDL_WINDOWS,            // Windows
      CSIDL_SYSTEM              // Windows\System
//    CSIDL_LOCAL_APPDATA,      // <user name>\Local Settings\Application Data (non roaming)
//    CSIDL_ALTSTARTUP,         // non localized startup
//    CSIDL_COMMON_ALTSTARTUP,  // non localized common startup
//    CSIDL_PROFILE             // UserProfile
   );

var
   CSIDL,
   Code   : Integer;
   sfPIDL : PItemIDList;
   sfPath : packed array[0..MAX_PATH] of Char;
begin
   { Under Win95, the CSIDL_COMMON_ special folders map to the corresponding non-commons }
   if IsWindowsNT then begin
      CSIDL := sfToCSIDL_WinNT[SpecialFolderType];
   end else begin
      CSIDL := sfToCSIDL_Win95[SpecialFolderType];
   end; { else }
   //These two CSIDLs fail on SHGetSpecialFolderLocation for no known reason
   case SpecialFolderType of
   sfWindows : begin
      Result := DUTGetWindowsDirectory;
      Exit;
   end; { sfWindows }
   sfSystem  : begin
      Result := DUTGetSystemDirectory;
      Exit;
   end; { sfSystem }
   else
      //No action
   end; { case }

   Code := SHGetSpecialFolderLocation(Application.Handle, CSIDL, sfPIDL);
   if NOERROR = Code then begin
      if SHGetPathFromIDList(sfPIDL, sfPath) then begin
         Result := String(sfPath);
      end else begin
         raise Exception.Create('GetSpecialFolderPath: SHGetPathFromIDList failed.');
      end; { else }
   end else begin
      raise Exception.Create('GetSpecialFolderPath: SHGetSpecialFolderLocation failed, code ' + IntToStr(Code));
   end; { else }
   AddTrailer(Result);
end; { GetSpecialFolderPath }

function FileProperties(const FileName : String) : Boolean;
{
   "Shells out" to the Properties window of the specified file or folder.
   Returns True if successfull, False otherwise.
}
var
   SEI : TShellExecuteInfo;
begin
   ZeroMemory(@SEI, SizeOf(TShellExecuteInfo));
   with SEI do begin
      { Size of structure }
      cbSize := SizeOf(TShellExecuteInfo);
      { Leave the process running;
        Create item ID list and invoke the application;
        No message box if error occurs }
      fMask  := SEE_MASK_NOCLOSEPROCESS or
                SEE_MASK_INVOKEIDLIST or
                SEE_MASK_FLAG_NO_UI;
      { See ShellExecuteEx for other possible flags }
      { Parent window of message boxes }
//    Wnd    := Handle;
      { Verb to execute (default is 'Open') }
      lpVerb := 'properties';
      { Supposedly a list of null-terminated strings;
        seems to only read the first string
        wild cards are NOT allowed }
      lpFile := PChar(FileName + #0);
      { Command line parameters }
//    lpParameters := '';
      { There ought to be a way to specify the tab... }
      { Working directory (Current directory used if unspecified) }
//    lpDirectory := '';
      { ShowWindow flags; ought to be zero since not an executable }
//    nShow  := SW_SHOWNORMAL; // SW_SHOW;

      { hInstApp will return an error code or the handle of the instance that was launched }
      { lpIDList ignored since fMask is not set to SEE_MASK_IDLIST }
      { lpClass ignored since fMask is not set to SEE_MASK_CLASSNAME }
      { hKeyClass ignored since fMask is not set to SEE_MASK_CLASSKEY }
      { dwHotKey ignored since fMask is not set to SEE_MASK_HOTKEY }
      { hIcon ignored since fMask is not set to SEE_MASK_ICON }
      { hProcess will return the handle of the launched application }
   end; { with }
   Result := ShellExecuteEx(@SEI);
end; { FileProperties }

function GetFileVersion(const FileName : String) : String;
{
   Obtains the specified file's Version string.
   Returns an empty string if the specified file
   isn't a Win32 executable (or DLL, etc.) or if it doesn't
   include a version string.
   For an executable to obtain it's own version info,
   it should pass in ParamStr(0).
}
var
   VISize,
   VIValueSize,
   Zero        : DWord;
   pVI,
   pVIValue    : Pointer;
begin
   Result := '';
   { Obtain size of version info structure }
   VISize := GetFileVersionInfoSize(PChar(FileName), Zero);
   if VISize = 0 then exit;
   { Allocate memory for the version info structure }
   { This could raise an EOutOfMemory exception }
   GetMem(pVI, VISize);
   try { pVI }
      if GetFileVersionInfo(PChar(FileName), 0, VISize, pVI) and
         VerQueryValue(pVI, sBackslash { root block }, pVIValue, VIValueSize) and
         (0 <> LongInt(VIValueSize)) then begin
         with TVSFixedFileInfo(pVIValue^) do begin
            Result :=                IntToStr(HiWord(dwFileVersionMS));
            Result := Result + '.' + IntToStr(LoWord(dwFileVersionMS));
            Result := Result + '.' + IntToStr(HiWord(dwFileVersionLS));
            Result := Result + '.' + IntToStr(LoWord(dwFileVersionLS));
         end; { with }
      end; { then }
   finally
      FreeMem(pVI);
   end; { try }
end; { GetFileVersion }

function GetFileNames(const Path : String; AttrReq, AttrExc : Integer; ExcludePath, Recurse : Boolean; ObjOption : TObjOption) : TStringList;
{
   Creates a sorted TStringList loaded with the files (and/or directories)
   found at Path (a fully qualified search pattern such as 'C:\*.*')
   Any directories returned will include the closing backslash.
   AttrReq specifies the attributes which *must* be satisfied,
   whilst AttrExc specifies those that must *not* appear.
   Setting both to zero is equivalent to asking for faAnyFile.
   If ExcludePath is set, the file names will exclude the Path part.
   Sub-directories are scanned only if Recurse is True (in which case,
   if ExcludePath is set *relative* file names will be returned).
   ObjOption indicates what to do with the TStringList's Object values.
      In all cases, the TSringList.Objects[] need to be typecast as Integer.
      ooIgnore   - Nothing.  All Object values will be Nil.
      ooFileSize - Object values are file sizes
      ooFileAttr - Object values are file attributes
      ooFileAge  - Object values are file ages (as per the FileAge function (q.v.))
                   TDateTime := FileDateToDateTime(Integer(TStringList.Objects[i]));
}
var
   F          : TSearchRec;
   S          : TStringList;
   PathPart,
   PathForName,
   FilterPart : String;
   i          : Integer;
begin
   Result := TStringList.Create;
   if 0 <> FindFirst(Path, faAnyFile, F) then Exit;
   try { F }
      PathPart   := ExtractFilePath(Path);
      if ExcludePath then begin
         PathForName := '';
      end else begin
         PathForName := PathPart;
      end; { else }
      FilterPart := ExtractFileName(Path);
      repeat
         if (F.Name = '.') or (F.Name = '..')
         then
            continue;
         if FlagIncluded(F.Attr, faDirectory) then AddTrailer(F.Name);
         if not(((AttrExc <> 0) and     FlagIncluded(F.Attr, AttrExc)) or
                ((AttrReq <> 0) and not FlagIncluded(F.Attr, AttrReq))
               )
         then begin
            case ObjOption of
            ooIgnore :
               Result.AddObject(PathForName + F.Name, Nil);
            ooFileSize :
               Result.AddObject(PathForName + F.Name, TObject(F.Size));
            ooFileAttr :
               Result.AddObject(PathForName + F.Name, TObject(F.Attr));
            ooFileAge :
               Result.AddObject(PathForName + F.Name, TObject(F.Time));
            end; { case }
         end; { then }
         if Recurse and FlagIncluded(F.Attr, faDirectory) then begin
            S := GetFileNames(PathPart + F.Name + FilterPart, AttrReq, AttrExc, ExcludePath, Recurse, ObjOption);
            try { S }
               if ExcludePath then begin
                  for i := 0 to (S.Count - 1) do begin
                     Result.AddObject(F.Name + S[i], S.Objects[i]);
                  end; { for }
               end else begin
                  Result.AddStrings(S); //Also copies Objects[]
               end; { else }
            finally
               S.Free;
            end; { try }
         end; { then }
      until 0 <> FindNext(F);
   finally
      FindClose(F);
   end; { try }
   Result.Sort;
end; { GetFileNames }

{ #################### Trash (Recycle Bin) Functions #################### }

function GetInfoRootPath(Drive : String) : String;
{
   Private function used by GetTrashContents and RestoreFromTrash.
   Returns the path to the Drive's recycled/recycler (sub)dir containing
   the Info or Info2 file.
   The concluding backslash is included.
}
var
   NTNetGetDCName     : function(Server, Domain : pWideChar; var DC : pWideChar) : DWORD {NET_API_STATUS}; stdcall;
   NTNetAPIBufferFree : function(lpBuffer : Pointer) : DWORD {NET_API_STATUS}; stdcall;

   PrimaryDomain : String;
   ErrMode       : Word;
   LibHandle     : THandle;
   pDomain       : PWideChar;
   cbSID         : Integer;
   SID           : PSID;
   peUse         : SID_NAME_USE;
   DomainName    : packed array [1..MAX_PATH] of Char;
   cbDomainName  : DWord;
begin
   Result := Drive + ':\' + RecycledDirStr + sBackslash;
   if not IsWindowsNT then Exit;

(*
   The location of the Info[2] file is different under WinNT:
   It appears as <Drive>:\Recycler\<some weird dir name>\Info
   According to Microsoft Knowledge Base Article Q171694, this
   weird dir name is a Security Identifier (SID); each logged-on
   user will thus see his own set of recycled files.
*)
   //Identify the WinNT intermediate directory
   {
   To obtain the correct SID, we must know the Primary Domain Controller's
   name (the domain the user logged onto).
   }
   PrimaryDomain := '';
   //We must dynamically bind to NetAPI32.dll since NetGetDCName and
   //NetAPIBufferFree are not available under Win95
   {
   Well, actually, NetGetDCName exists in 16-bit NetAPI.dll but its
   syntax is different: instead of passing a variable that receives
   a buffer address that must later be freed by NetAPIBufferFree, you
   pass it a buffer and its length.  NetGetDCName then writes the
   primary domain controller's name to the buffer.
   The return codes are the same.
   Example:
   cbDomain := MAX_PATH;
   GetMem(pDomain, cbDomain);
   if NERR_SUCCESS <> NetGetDCName(Nil, Nil, pDomain, cbDomain) then
   }
   ErrMode := SetErrorMode(SEM_NOOPENFILEERRORBOX);
   LibHandle := LoadLibrary('NetAPI32.dll');
   SetErrorMode(ErrMode);
   if LibHandle = 0 then raise Exception.Create('Unable to load library NetAPI32.dll');
   try { LibHandle }
      @NTNetGetDCName     := GetProcAddress(LibHandle, 'NetGetDCName');
      @NTNetAPIBufferFree := GetProcAddress(LibHandle, 'NetApiBufferFree');
      SetLastError(NTNetGetDCName(Nil, Nil, pDomain));
      if NERR_SUCCESS <> GetLastError then
         //I'm not sure SysErrorMessage will work since the NERR_ codes are in the 2100..2999 range
         raise Exception.CreateFmt('%s (Error code %d)', [SysErrorMessage(GetLastError), GetLastError]);
      try { pDomain }
         PrimaryDomain := WideCharToString(pDomain);
      finally
         NTNetAPIBufferFree(pDomain);
      end; { try }
   finally
      FreeLibrary(LibHandle);
   end; { try }

   //Allocate SID
   cbSID := 1024;
   cbDomainName := MAX_PATH;
   GetMem(SID, cbSID);
   try { SID }
      //LookupAccountName isn't supposed to be available under Win95,
      //according to Delphi 3's Win32.hlp.  However, my Win95 system's
      //AdvAPI32.dll does include it, so...
      //if not LookupAccountName(PChar(DUTGetSecurityProvider), PChar(DUTGetUserName), SID, cbSID, @DomainName, cbDomainName, peUse) then begin
      if not LookupAccountName(PChar(PrimaryDomain), PChar(DUTGetUserName), SID, cbSID, @DomainName, cbDomainName, peUse) then
         raise Exception.Create('Unable to obtain SID');
      //peUse ought to be SIDTypeUser
      Result := Drive + ':\' + RecycledDirNTStr + sBackslash + SIDToStr(SID) + sBackslash;
   finally
      FreeMem(SID); //FreeSID(SID);
   end; { try }
end; { GetInfoRootPath }

function GetTrashContents(DeletionDateTimes, FileSizes : TList) : TStringList;
{
   Returns a TStringList loaded with the fully qualified names of the files
   currently in the Trash (a.k.a Recycle Bin).
   The TStringList could be empty.
   It is the responsability of the calling application to Free the TStringList.
   If either or both of DeletionDateTimes and FileSizes are non-Nil, they will be
   loaded (in the same sequence as the TStringList) with the files' deletion
   times (as pointers to TDateTime objects) and sizes (as pointers to Integers).
   It is the responsability of the calling application to Free the pointed
   TDateTime and/or Integer objects.
}
var
   SM           : IMalloc;
   SF,
   TF           : IShellFolder;
   ItemPIDL,
   TrashPIDL    : PItemIDList;
   EL           : IEnumIDList;
   Fetched      : ULONG;
   Drive        : TDriveLetter;
   NumFile,
   NumSought    : Integer;
   InfosPos     : LongInt;
   FilePath,
   InfoPath     : packed array [1..MAX_PATH] of Char; //Must be precisely MAX_PATH bytes
   InfoRootPath : String;
   Infos        : array[TDriveLetter] of TFileStream;
   InfoNum      : array[TDriveLetter] of Integer;
   DeletedOn,
   DeletedOn2        : TFileTime;
   FileSize,
   DOSDelDateTime    : Integer;
   pDeletionDateTime : ^TDateTime;
   pFileSize         : ^Integer;
begin
   Result := TStringList.Create;
   try { Result }
      for Drive := Low(TDriveLetter) to High(TDriveLetter) do Infos[Drive] := Nil;
      //We'll need IMalloc to free the returned PIDLs
      OLECheck(SHGetMalloc(SM));
      //Obtain IShellFolder for desktop
      OLECheck(SHGetDesktopFolder(SF));
      //Obtain Trash
      OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
      try { TrashPIDL }
         //Bind to the Trash
         OLECheck(SF.BindToObject(TrashPIDL, Nil, IID_IShellFolder, Pointer(TF)));
      finally
         //We no longer need the TrashPIDL
         SM.Free(TrashPIDL);
      end; { try }
      //Obtain IEnumIDList of Trash contents
      OLECheck(TF.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EL));
      //Fetch first item in Trash
      //Note that S_FALSE is NOT considered an error by OLECheck
      OLECheck(EL.Next(1, ItemPIDL, Fetched));
      try { Infos[] }
         while Fetched > 0 do begin
            try { ItemPIDL }
               {
               At this point, calling SHGetPathFromIDList(ItemPIDL, ...) returns the
               true file title ('D'<OriginalDriveLetter><Number>.<extension>)
               but always returns the path as 'C:\Windows\Desktop\' since the
               ItemPIDL isn't relative to the Desktop (as SHGetPathFromIDList
               expects) but rather relative to the Trash.
               We'll use the file title information to locate the proper
               'Recycled' directory and read it's Info (or Info2) file...
               }
               OLECheckBoolean(SHGetPathFromIDList(ItemPIDL, @InfoPath));
            finally
               //We no longer need this PIDL
               SM.Free(ItemPIDL);
            end; { try }

            //InfoPath is 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>.<extension>
            //         or 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>
            //            'C:\Documents and Settings\<username>\Desktop\D'<OriginalDriveLetter><Number>[.<extension>] (Win 2000)
            //The item may be a directory.
            //Extract drive letter from trash file title
            //Win 2k actually stores the drive letter in lower case
            Drive := UpperCase(ExtractFileName(String(PChar(@InfoPath))))[2];
            //Extract sequence number from trash file title
            NumSought := StrToInt(Copy(ExtractFileTitle(String(PChar(@InfoPath))), 3, MAX_PATH));
            //Obtain the path to the Info[2] file
            InfoRootPath := GetInfoRootPath(Drive);

            //If we haven't opened the corresponding drive's Info file yet, do so
            if not Assigned(Infos[Drive]) then begin
               try { Info2 vs Info }
                  Infos[Drive] := TFileStream.Create(InfoRootPath + 'Info2', fmOpenRead or fmShareDenyNone);
               except
                  Infos[Drive] := TFileStream.Create(InfoRootPath + 'Info',  fmOpenRead or fmShareDenyNone);
               end; { try }
               //Skip 20-byte header
               Infos[Drive].Seek(20, soFromBeginning);
            end; { then }
            //Look ahead to the next record's FileSequenceNumber
            InfosPos := Infos[Drive].Position;
            Infos[Drive].Seek(MAX_PATH, soFromCurrent);
            Infos[Drive].Read(NumFile, SizeOf(NumFile));
            InfoNum[Drive] := NumFile;
            //Rewind
            Infos[Drive].Position := InfosPos;

            //Read the file info
            //The EL returns the files in the same sequence as they appear in
            //the Recycled\Info file; however, there may be gaps in the latter.
            //Skip gaps
            while InfoNum[Drive] < NumSought do begin
               //Skip a record
               Infos[Drive].Seek(MAX_PATH*(1 + 2*Ord(IsWindowsNT)) + 20, soFromCurrent);
               //Look ahead to the next record's FileSequenceNumber
               InfosPos := Infos[Drive].Position;
               Infos[Drive].Seek(MAX_PATH, soFromCurrent);
               Infos[Drive].Read(NumFile, SizeOf(NumFile));
               InfoNum[Drive] := NumFile;
               //Rewind
               Infos[Drive].Position := InfosPos;
            end; { while }

            //Read the file's fully qualified original path and name
            Infos[Drive].Read(FilePath, MAX_PATH);
            //The String(PChar()) typecast sets the Length of the string to its correct value
            Result.Add(String(PChar(@FilePath)));

            //Skip 8 bytes:
            // FileSequenceNumber : Integer;
            // DriveLetter        : Integer; { 'A' = 0, 'B' = 1... }
            Infos[Drive].Seek(8, soFromCurrent);

            //Get deletion time
            //DeletedOn : TFileTime is the UTC time of deletion
            Infos[Drive].Read(DeletedOn, SizeOf(TDateTime));
            if Assigned(DeletionDateTimes) then begin
               //Convert from UTC to local time
               DeletedOn2 := DeletedOn;
               OLECheckBoolean(FileTimeToLocalFileTime(DeletedOn2, DeletedOn));
               //Convert to DOS date-time value
               OLECheckBoolean(FileTimeToDosDateTime(DeletedOn, LongRec(DOSDelDateTime).Hi, LongRec(DOSDelDateTime).Lo));
               GetMem(pDeletionDateTime, SizeOf(TDateTime));
               try { pDeletionDateTime }
                  //Convert to TDateTime
                  pDeletionDateTime^ := FileDateToDateTime(DOSDelDateTime);
                  DeletionDateTimes.Add(pDeletionDateTime);
               except
                  FreeMem(pDeletionDateTime);
                  raise;
               end; { try }
            end; { then }

            //Get file size
            {
            This obtains the file's footprint size; that is,
            the number of bytes it occupies on the drive
            (a whole multiple of the drive's cluster size)
            }
            //We need to do the Read just to advance the file cursor
            Infos[Drive].Read(FileSize, SizeOf(Integer));
            {
            This obtains the file's true size; that is,
            the number of bytes it contains
            }
            FileSize := DUTGetFileSize(Drive + ':\' + RecycledDirStr + sBackslash + ExtractFileName(String(PChar(@InfoPath))));
            if Assigned(FileSizes) then begin
               GetMem(pFileSize, SizeOf(Integer));
               try { pFileSize }
                  pFileSize^ := FileSize;
                  FileSizes.Add(pFileSize);
               except
                  FreeMem(pFileSize);
                  raise;
               end; { try }
            end; { then }

            //Skip Unicode file name
            if IsWindowsNT then Infos[Drive].Seek(MAX_PATH*2, soFromCurrent);

            //Next item
            OLECheck(EL.Next(1, ItemPIDL, Fetched));
         end; { while }
      finally
         //Cleanup --close all Recycled\Info files
         for Drive := Low(TDriveLetter) to High(TDriveLetter) do Infos[Drive].Free;
      end; { try }
   except
      Result.Free;
//    Result := Nil;
      raise;
   end; { try }
end; { GetTrashContents }

function RestoreFromTrash(FileName : String) : Boolean;
{
   Attempts to Restore the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}
const
   CmdID_First = 0;
var
   SM           : IMalloc;
   TF,
   SF           : IShellFolder;
   ItemPIDL,
   TrashPIDL    : PItemIDList;
   EL           : IEnumIDList;
   Fetched      : ULONG;
   InfoNum,
   NumSought    : Integer;
   InfosPos     : LongInt;
   InfoPath,
   FilePath     : packed array [1..MAX_PATH] of Char; //Must be precisely MAX_PATH bytes
   InfoRootPath : String;
   Info         : TFileStream;
   ContextMenu  : IContextMenu;
   Popup        : HMenu;
   ICI          : TCMInvokeCommandInfo;
   MII          : TMenuItemInfo;
   hObject      : THandle;
begin
   Result := False;
   if FileName = '' then Exit;
   OLECheck(SHGetMalloc(SM));;
   //Obtain IShellFolder for desktop
   OLECheck(SHGetDesktopFolder(SF));
   //Obtain Trash
   OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
   try { TrashPIDL }
      //Obtain the path to the Info[2] file
      InfoRootPath := GetInfoRootPath(FileName[1]);

      //Open the corresponding Recycled\Info file
      try { Info2 vs Info }
         Info := TFileStream.Create(InfoRootPath + 'Info2', fmOpenRead or fmShareDenyNone);
      except
         Info := TFileStream.Create(InfoRootPath + 'Info',  fmOpenRead or fmShareDenyNone);
      end; { try }

      try { Info }
         //Skip 20-byte header
         Info.Seek(20, soFromBeginning);

         //Bind to the Trash
         OLECheck(SF.BindToObject(TrashPIDL, Nil, IID_IShellFolder, Pointer(TF)));
         //Obtain IEnumIDList of Trash contents
         OLECheck(TF.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EL));
         //Fetch first item in Trash
         //Note that S_FALSE is NOT considered an error by OLECheck
         OLECheck(EL.Next(1, ItemPIDL, Fetched));
         try { ItemPIDL }
            while Fetched > 0 do begin
               OLECheckBoolean(SHGetPathFromIDList(ItemPIDL, @InfoPath));
               //InfoPath is 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>.<extension>
               //Extract drive letter from trash file title
               //If not the FileName's drive, move on to next item in Trash
               if 0 = ANSICompareText(FileName[1], ExtractFileName(String(PChar(@InfoPath)))[2]) then begin
                  //Extract sequence number from trash file title
                  NumSought := StrToInt(Copy(ExtractFileTitle(String(PChar(@InfoPath))), 3, MAX_PATH));

                  //Look ahead to the record's FileSequenceNumber
                  InfosPos := Info.Position;
                  Info.Seek(MAX_PATH, soFromCurrent);
                  Info.Read(InfoNum, SizeOf(InfoNum));
                  //Rewind
                  Info.Position := InfosPos;

                  //Read the file info
                  //The EL returns the files in the same sequence as they appear in
                  //the Recycled\Info file; however, there may be gaps in the latter.
                  //Skip gaps
                  while InfoNum < NumSought do begin
                     //Skip a record
                     Info.Seek(MAX_PATH*(1 + 2*Ord(IsWindowsNT)) + 20, soFromCurrent);
                     //Look ahead to the next record's FileSequenceNumber
                     InfosPos := Info.Position;
                     Info.Seek(MAX_PATH, soFromCurrent);
                     Info.Read(InfoNum, SizeOf(InfoNum));
                     //Rewind
                     Info.Position := InfosPos;
                  end; { while }

                  //Read the file's original fully qualified name
                  Info.Read(FilePath, MAX_PATH);
                  //Skip the remaining 20 bytes
                  Info.Seek(20, soFromCurrent);
                  //Skip Unicode file name
                  if IsWindowsNT then Info.Seek(MAX_PATH*2, soFromCurrent);

                  //The String(PChar()) typecast sets the Length of the string to its correct value
                  if 0 = ANSICompareText(FileName, String(PChar(@FilePath))) then break;
               end; { then }
               //Next item in Trash
               SM.Free(ItemPIDL); ItemPIDL := Nil;
               OLECheck(EL.Next(1, ItemPIDL, Fetched));
            end; { while }
            //Did we find FileName?
            //Note that this failure mode is not an Exception
            if Fetched = 0 then Exit;

            //We found the FileName
            //Obtain IContextMenu
            OLECheck(TF.GetUIObjectOf(Application.Handle, 1, ItemPIDL, IID_IContextMenu, Nil, Pointer(ContextMenu)));
         finally
            //PIDL no longer needed
            //However, it may have already been Freed in two ways:
            // 1) We did not find the FileName; or
            // 2) EL.Next failed
            if Assigned(ItemPIDL) then SM.Free(ItemPIDL);
         end; { try }
      finally
         Info.Free;
      end; { try }

      //Populate menu
      Popup := CreatePopupMenu;
      OLECheckBoolean(Popup <> 0);
      try { Popup }
         OLECheck(ContextMenu.QueryContextMenu(Popup, 0 {insertion index}, CmdID_First, $7FFF {CmdID_Last}, CMF_EXPLORE));
         //Identify the "Restore" command (unfortunately, the Shell supplies no Verb!)
         OLECheckBoolean(GetMenuItemInfoByCaption(Popup, RestoreCmdStr, MII));
         //Invoke the command
         FillChar(ICI, SizeOf(TCMInvokeCommandInfo), #0);
         with ICI do begin
            cbSize := SizeOf(TCMInvokeCommandInfo);
            //Does NOT prevent confirmation prompt (for read-only files and such)
//          fMask := CMIC_MASK_FLAG_NO_UI;
            hWnd := Application.Handle;
            nShow := SW_SHOWNORMAL;
            lpVerb := PChar(MII.wID - CmdID_First);
         end; { with }
         OLECheck(ContextMenu.InvokeCommand(ICI));
         Result := True;

         //It seems the shell dawdles a bit before updating the Info files
         SHChangeNotify(SHCNE_UPDATEDIR, SHCNF_FLUSH or SHCNF_IDLIST, TrashPIDL, Nil);
         hObject := FindFirstChangeNotification(PChar(FileName[1] + ':\' + RecycledDirStr), False, FILE_NOTIFY_CHANGE_LAST_WRITE);
         OLECheckBoolean(INVALID_HANDLE_VALUE <> hObject);
         try { hObject }
            //We do not want to throw an exception if the user is faced with a confirmation prompt
//          OLECheckBoolean(WAIT_OBJECT_0 = WaitForSingleObject(hObject, 500 { ms }));
            WaitForSingleObject(hObject, 500 { ms });
         finally
            OLECheckBoolean(FindCloseChangeNotification(hObject));
         end; { try }
      finally
         OLECheckBoolean(DestroyMenu(Popup));
      end; { try }
   finally
      SM.Free(TrashPIDL);
   end; { try }
end; { RestoreFromTrash }

function DeleteFromTrash(FileName : String) : Boolean;
{
   Attempts to Delete the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}
const
   CmdID_First = 0;
var
   SM           : IMalloc;
   TF,
   SF           : IShellFolder;
   ItemPIDL,
   TrashPIDL    : PItemIDList;
   EL           : IEnumIDList;
   Fetched      : ULONG;
   InfoNum,
   NumSought    : Integer;
   InfosPos     : LongInt;
   InfoPath,
   FilePath     : packed array [1..MAX_PATH] of Char; //Must be precisely MAX_PATH bytes
   InfoRootPath : String;
   Info         : TFileStream;
   ContextMenu  : IContextMenu;
   Popup        : HMenu;
   ICI          : TCMInvokeCommandInfo;
   { This is one departure from the RestoreFromTrash code }
// MII          : TMenuItemInfo;
   hObject      : THandle;
begin
   Result := False;
   if FileName = '' then Exit;
   OLECheck(SHGetMalloc(SM));;
   //Obtain IShellFolder for desktop
   OLECheck(SHGetDesktopFolder(SF));
   //Obtain Trash
   OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
   try { TrashPIDL }
      //Obtain the path to the Info[2] file
      InfoRootPath := GetInfoRootPath(FileName[1]);

      //Open the corresponding Recycled\Info file
      try { Info2 vs Info }
         Info := TFileStream.Create(InfoRootPath + 'Info2', fmOpenRead or fmShareDenyNone);
      except
         Info := TFileStream.Create(InfoRootPath + 'Info',  fmOpenRead or fmShareDenyNone);
      end; { try }

      try { Info }
         //Skip 20-byte header
         Info.Seek(20, soFromBeginning);

         //Bind to the Trash
         OLECheck(SF.BindToObject(TrashPIDL, Nil, IID_IShellFolder, Pointer(TF)));
         //Obtain IEnumIDList of Trash contents
         OLECheck(TF.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EL));
         //Fetch first item in Trash
         //Note that S_FALSE is NOT considered an error by OLECheck
         OLECheck(EL.Next(1, ItemPIDL, Fetched));
         try { ItemPIDL }
            while Fetched > 0 do begin
               OLECheckBoolean(SHGetPathFromIDList(ItemPIDL, @InfoPath));
               //InfoPath is 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>.<extension>
               //Extract drive letter from trash file title
               //If not the FileName's drive, move on to next item in Trash
               if 0 = ANSICompareText(FileName[1], ExtractFileName(String(PChar(@InfoPath)))[2]) then begin
                  //Extract sequence number from trash file title
                  NumSought := StrToInt(Copy(ExtractFileTitle(String(PChar(@InfoPath))), 3, MAX_PATH));

                  //Look ahead to the next record's FileSequenceNumber
                  InfosPos := Info.Position;
                  Info.Seek(MAX_PATH, soFromCurrent);
                  Info.Read(InfoNum, SizeOf(InfoNum));
                  //Rewind
                  Info.Position := InfosPos;

                  //Read the file info
                  //The EL returns the files in the same sequence as they appear in
                  //the Recycled\Info file; however, there may be gaps in the latter.
                  //Skip gaps
                  while InfoNum < NumSought do begin
                     Info.Seek(MAX_PATH*(1 + 2*Ord(IsWindowsNT)) + 20, soFromCurrent);
                     //Look ahead to the next record's FileSequenceNumber
                     InfosPos := Info.Position;
                     Info.Seek(MAX_PATH, soFromCurrent);
                     Info.Read(InfoNum, SizeOf(InfoNum));
                     //Rewind
                     Info.Position := InfosPos;
                  end; { while }

                  //Read the file's original fully qualified name
                  Info.Read(FilePath, MAX_PATH);
                  //Skip the remaining 20 bytes
                  Info.Seek(20, soFromCurrent);
                  //Skip Unicode file name
                  if IsWindowsNT then Info.Seek(MAX_PATH*2, soFromCurrent);

                  //The String(PChar()) typecast sets the Length of the string to its correct value
                  if 0 = ANSICompareText(FileName, String(PChar(@FilePath))) then break;
               end; { then }
               //Next item in Trash
               SM.Free(ItemPIDL); ItemPIDL := Nil;
               OLECheck(EL.Next(1, ItemPIDL, Fetched));
            end; { while }
            //Did we find FileName?
            //Note that this failure mode is not an Exception
            if Fetched = 0 then Exit;

            //We found the FileName
            //Obtain IContextMenu
            OLECheck(TF.GetUIObjectOf(Application.Handle, 1, ItemPIDL, IID_IContextMenu, Nil, Pointer(ContextMenu)));
         finally
            //PIDL no longer needed
            //However, it may have already been Freed in two ways:
            // 1) We did not find the FileName; or
            // 2) EL.Next failed
            if Assigned(ItemPIDL) then SM.Free(ItemPIDL);
         end; { try }
      finally
         Info.Free;
      end; { try }

      //Populate menu
      Popup := CreatePopupMenu;
      OLECheckBoolean(Popup <> 0);
      try { Popup }
         OLECheck(ContextMenu.QueryContextMenu(Popup, 0 {insertion index}, CmdID_First, $7FFF {CmdID_Last}, CMF_EXPLORE));
         { This is another departure from the RestoreFromTrash code }
         //No need to identify the "Delete" command since the Shell does supply a Verb in this case
//       OLECheckBoolean(GetMenuItemInfoByCaption(Popup, DeleteCmdStr, MII));
         //Invoke the command
         FillChar(ICI, SizeOf(TCMInvokeCommandInfo), #0);
         with ICI do begin
            cbSize := SizeOf(TCMInvokeCommandInfo);
            //Does NOT prevent confirmation prompt (for read-only files and such)
//          fMask := CMIC_MASK_FLAG_NO_UI;
            hWnd := Application.Handle;
            nShow := SW_SHOWNORMAL;
            { This is the last departure from the RestoreFromTrash code }
//          lpVerb := PChar(MII.wID - CmdID_First);
            lpVerb := PChar(CMDSTR_DELETE);
         end; { with }
         OLECheck(ContextMenu.InvokeCommand(ICI));
         Result := True;

         //It seems the shell dawdles a bit before updating the Info files
         SHChangeNotify(SHCNE_UPDATEDIR, SHCNF_FLUSH or SHCNF_IDLIST, TrashPIDL, Nil);
         hObject := FindFirstChangeNotification(PChar(FileName[1] + ':\' + RecycledDirStr), False, FILE_NOTIFY_CHANGE_LAST_WRITE);
         OLECheckBoolean(INVALID_HANDLE_VALUE <> hObject);
         try { hObject }
            //We do not want to throw an exception if the user is faced with a confirmation prompt
//          OLECheckBoolean(WAIT_OBJECT_0 = WaitForSingleObject(hObject, 500 { ms }));
            WaitForSingleObject(hObject, 500 { ms });
         finally
            OLECheckBoolean(FindCloseChangeNotification(hObject));
         end; { try }
      finally
         OLECheckBoolean(DestroyMenu(Popup));
      end; { try }
   finally
      SM.Free(TrashPIDL);
   end; { try }
end; { DeleteFromTrash }

function EmptyTrash : Boolean;
{
   Empties the Trash (a.k.a Recycle Bin).
   Returns False if the trash was already empty.
}
const
   CmdID_First = 0;
var
   SM          : IMalloc;
   SF          : IShellFolder;
   TrashPIDL   : PItemIDList;
   ContextMenu : IContextMenu;
   Popup       : HMenu;
   ICI         : TCMInvokeCommandInfo;
   MII         : TMenuItemInfo;
begin
   Result := False;
   OLECheck(SHGetMalloc(SM));
   //Obtain IShellFolder for desktop
   OLECheck(SHGetDesktopFolder(SF));
   //Obtain Trash
   OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
   try { TrashPIDL }
      //Obtain IContextMenu
      OLECheck(SF.GetUIObjectOf(Application.Handle, 1, TrashPIDL, IID_IContextMenu, Nil, Pointer(ContextMenu)));
      //Populate menu
      Popup := CreatePopupMenu;
      OLECheckBoolean(Popup <> 0);
      try { Popup }
         OLECheck(ContextMenu.QueryContextMenu(Popup, 0 {insertion index}, CmdID_First, $7FFF {CmdID_Last}, CMF_EXPLORE));
         //Identify the "Empty Recycle Bin" command (unfortunately, the Shell supplies no Verb!)
         OLECheckBoolean(GetMenuItemInfoByCaption(Popup, EmptyTrashCmdStr, MII));
         //Enabled?
         if FlagIncluded(MII.fState, MFS_ENABLED) then begin
            //Invoke the command
            FillChar(ICI, SizeOf(TCMInvokeCommandInfo), #0);
            with ICI do begin
               cbSize := SizeOf(TCMInvokeCommandInfo);
               hWnd := Application.Handle;
               nShow := SW_SHOWNORMAL;
               lpVerb := PChar(MII.wID - CmdID_First);
            end; { with }
            OLECheck(ContextMenu.InvokeCommand(ICI));
            Result := True;
            //Wait for the shell to update its Info files
            SHChangeNotify(SHCNE_UPDATEDIR, SHCNF_FLUSH or SHCNF_IDLIST, TrashPIDL, Nil);
         end; { then }
      finally
         OLECheckBoolean(DestroyMenu(Popup));
      end; { try }
   finally
      SM.Free(TrashPIDL);
   end; { try }
end; { EmptyTrash }

end. { DUTUtilsFiles }
