unit DUTUtilsMisc;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   22 June 1999
    1998-1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   Miscellaneous utilities.
*)
interface

type
   //A 64-bit integer
   HugeInt = packed record
   Lo, Hi : Integer; //LongInt - 32 bits each
   end; { HugeInt }

{ #################### Integer Flag Manipulation #################### }

function FlagIncluded(const Flags, Flag : Integer) : Boolean;
{
   Returns True if the bit pattern Flag is included in Flags.
}

{ #################### Calendar (Date/Time) Functions #################### }

function CalendarWeek(ADate : TDateTime) : Integer;
{
   Calculates calendar week (1..52) assuming:
   - First week includes 1 January.
   - Week starts on Sunday (as per DayOfWeek function).
}

implementation

uses
   SysUtils; { DecodeDate, etc. }

{ #################### Integer Flag Manipulation #################### }

function FlagIncluded(const Flags, Flag : Integer) : Boolean;
{
   Returns True if the bit pattern Flag is included in Flags.
}
begin
   Result := Flag = (Flags and Flag);
end; { FlagIncluded }

{ #################### Calendar (Date/Time) Functions #################### }

function CalendarWeek(ADate : TDateTime) : Integer;
{
   Calculates calendar week (1..52) assuming:
   - First week includes 1 January.
   - Week starts on Sunday (as per DayOfWeek function).
}
var
   FirstOfYear : TDateTime;
   Day,
   Month,
   Year        : Word;
begin
   DecodeDate(ADate, Year, Month, Day);
   FirstOfYear := EncodeDate(Year, 1, 1);
   //ADate - FirstOfYear gives the number of days elapsed since 1 January
   //We must offset by 0..6 based on which DayOfWeek 1 January fell on.
   Result := 1 + (Trunc((DayOfWeek(FirstOfYear) - 1) + ADate - FirstOfYear) div 7);
end; { CalendarWeek }
(*
function CalendarWeek(ADate : TDateTime) : integer;
{
   Calculates calendar week assuming:
   - Monday is the 1st day of the week.
   - Week one is the first week of the year to include a Thursday.
   If Result is 0, then ADate falls on the last week of the previous year.
   (This may happen for 1-3 January)

   Based on code by TeamB's Ralph Friedman.
}
var
   FirstOfYear : TDateTime;
   DayOne,
   MonthOne,
   Day,
   Month,
   Year        : Word;
begin
   DecodeDate(ADate, Year, Month, Day);

   //DayOne is the day on which week one begins
   case DayOfWeek(EncodeDate(Year, 1, 1)) of
   1 : DayOne := 2;  // Sunday  -- Week one begins 2 Jan
   2 : DayOne := 1;  // Monday
   3 : DayOne := 31; // Tuesday -- Week one began 31 Dec
   4 : DayOne := 30; // Wednesday
   5 : DayOne := 29; // Thursday
   6 : DayOne := 4;  // Friday  -- Week one begins 4 Jan
   7 : DayOne := 3;  // Saturday
   end; { case }

   //MonthOne is the month on which week one begins
   if DayOne > 4 then begin
      Dec(Year);
      MonthOne := 12; //December
   end else begin
      MonthOne := 1;  //January
   end; { else }

   //FirstOfYear is a Monday
   FirstOfYear := EncodeDate(Year, MonthOne, DayOne);

   if ADate < FirstOfYear then begin
      Result := 0;
   end else begin
      Result := 1 + (Trunc(ADate - FirstOfYear) div 7);
   end; { else }
end; { CalendarWeek }
*)

end. { DUTUtilsMisc }
