unit DUTUtilsReg;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   22 June 1999
    1998-1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   Registry utilities.
   Includes a public TheRegistry TRegistry object.
*)
interface

uses
   Windows,  { HKey }
   Registry; { TRegistry }

{ #################### INI Registry Manipulation #################### }

function GetINIInteger(Root : HKey; Path, ValueName : String; Default : Integer) : Integer;
{
   Obtains the integer Value from Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   Returns Default if not found.
}

function GetINIString(Root : HKey; Path, ValueName : String; Default : String) : String;
{
   Obtains the string Value from Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   Returns Default if not found.
}

procedure SetINIInteger(Root : HKey; Path, ValueName : String; Value : Integer);
{
   Sets the integer Value of Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   The key and value are created if necessary.
}

procedure SetINIString(Root : HKey; Path, ValueName : String; Value : String);
{
   Sets the string Value of Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   The key and value are created if necessary.
}

var
   TheRegistry : TRegistry;

implementation

{ #################### INI File Manipulation #################### }

function GetINIInteger(Root : HKey; Path, ValueName : String; Default : Integer) : Integer;
{
   Obtains the integer Value from Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   Returns Default if not found.
}
begin
   Result := Default;
   TheRegistry.RootKey := Root;
   if not TheRegistry.KeyExists(Path) then exit;
   TheRegistry.OpenKey(Path, False);
   if not TheRegistry.ValueExists(ValueName) then exit;
   Result := TheRegistry.ReadInteger(ValueName);
end; { GetINIInteger }

function GetINIString(Root : HKey; Path, ValueName : String; Default : String) : String;
{
   Obtains the string Value from Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   Returns Default if not found.
}
begin
   Result := Default;
   TheRegistry.RootKey := Root;
   if not TheRegistry.KeyExists(Path) then exit;
   TheRegistry.OpenKey(Path, False);
   if not TheRegistry.ValueExists(ValueName) then exit;
   Result := TheRegistry.ReadString(ValueName);
end; { GetINIString }

procedure SetINIInteger(Root : HKey; Path, ValueName : String; Value : Integer);
{
   Sets the integer Value of Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   The key and value are created if necessary.
}
begin
   TheRegistry.RootKey := Root;
   TheRegistry.OpenKey(Path, True);
   TheRegistry.WriteInteger(ValueName, Value);
end; { SetINIInteger }

procedure SetINIString(Root : HKey; Path, ValueName : String; Value : String);
{
   Sets the string Value of Root\Path : ValueName.
   HKEY_LOCAL_MACHINE is recommended for Root;
   '\Software\<Company>\<Application>\<Category>' is recommended for Path.
   The key and value are created if necessary.
}
begin
   TheRegistry.RootKey := Root;
   TheRegistry.OpenKey(Path, True);
   TheRegistry.WriteString(ValueName, Value);
end; { SetINIString }

initialization

   TheRegistry := TRegistry.Create;

finalization

   TheRegistry.Free;

end. { DUTUtilsReg }
