unit DUTUtilsSys;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   23 August 1999
   Revised 30 January 2000
    1998-2000 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   System utilities.
*)
interface

uses
   Windows,  { LCID, etc. }
   Classes,  { TShiftState }
   Dialogs,  { CreateMessageDialog, TMsgDlgType }
   Forms;    { TForm, TApplication }

const
   MK_ALT =    $20; //Add-on to windows' MK_ constants
   MK_DOUBLE = $40; //Add-on to windows' MK_ constants
   MK_WIN =    $1F; //Used to filter out the add-ons

const
   //WinNT declarations from WinNT.h
   SID_REVISION                    =  1; // Current revision level
   SID_MAX_SUB_AUTHORITIES         = 15;
// SID_RECOMMENDED_SUB_AUTHORITIES =  1; // Will change to around 6 in a future release
   //SID_Name_Use enumeration
// SIDTypeComputer = 9;
   //There are a whole bunch of standard SIDs and stuff left in WinNT.h

{ #################### WinNT Security ID (SID) Functions #################### }

function SIDToStr(SID : PSID) : String;
{
   Converts the SID into its string representation.
   An exception is raised if the SID is invalid.
   The string will be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*
   SID_REVISION is a single byte
   SID_IDENTIFIER_AUTHORITY is 0x000000000000
   SIDSubAuthorities are DWords

   We need this function because ConvertSIDToStringSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}

function StrToSID(const SIDName : String) : PSID;
{
   Allocates and fills in an SID based on the string SIDName.
   An exception is raised (and no memory allocated) if the
   SID creation fails.
   The SIDName should be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*
   The caller is responsible for calling FreeSID to free the PSID.

   We need this function because ConvertStringSIDToSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}

{ #################### Message Utilities #################### }

function FlashMessagePanel(const Msg : String; AType : TMsgDlgType; HelpCtx : Longint; X, Y : Integer; const HelpFileName : String; MilliSeconds : Integer) : TForm;
{
   Flashes a message panel (a button-less dialog) in non-modal fashion.
   Msg          is the message displayed on the panel
   AType        is the TMsgDlgType (as per the MessageDlg functions)
   HelpCtx      is an optional Help context (zero for none)
   X, Y         are optional Left and Top positions (-1 for default)
   HelpFileName is an optional Help file name ('' for default)
   Milliseconds is how long to display the panel

   If Milliseconds is non-zero, FlashMessagePanel waits for the specified time and then
   disposes of the message panel; the function returns Nil.  The window will have no
   BorderIcons in that case (no close box).
   If Milliseconds is zero (or less), FlashMessagePanel throws the message panel up
   and then returns its TForm reference.  It is up to the caller to Free it.  The
   window will have a close box.
}

{ #################### Spawning Utilities #################### }

function RunItAndWait(AppCommand : String) : LongInt;
{
   "Shells out" to the AppCommand (which must be a DOS or Windows command line)
   and waits for it to terminate.
   The exit code (or an error code) is returned.
   For the 16-bit Setup.exe stub app used by Install Shield Express,
   this won't work.  I'm at a loss to explain why.
}

{ #################### System Info #################### }

function DUTGetUserName : String;
{
   Obtains the User Name.
   Returns the '.Default' string if no "user" has logged on.
}

function DUTGetNetworkUserName : String;
{
   Obtains the network User Name.
   Returns an empty string if there is no network or
   the user did not log on.
   Raises an exception otherwise.
}

function IsWindows95 : Boolean;
{
   Returns True if the OS is Windows 95 (or 98)
}

function IsWindowsNT : Boolean;
{
   Returns True if the OS is Windows NT
}

function GetLocaleNumberFmt(Locale : LCID; var NumSepStr : String; var ThsndsStr : String) : TNumberFmt;
{
   Obtains the TNumberFmt record from the system for the specified Locale;
   NumSepStr and ThsndsStr must be supplied as they'll be pointed to by
   the TNumberFmt.
}

function GetLocaleCurrencyFmt(Locale : LCID; var NumSepStr : String; var ThsndsStr : String; var CrrncyStr : String) : TCurrencyFmt;
{
   Obtains the TCurrencyFmt record from the system for the specified Locale;
   NumSepStr, ThsndsStr and CrrncyStr must be supplied as they'll be pointed
   to by the TCurrencyFmt.
}

function GetDOSEnvironmentVar(const VarName : String) : String;
{
   Returns the DOS Environment variable VarName (such as PATH or TEMP).
   If VarName isn't defined, returns an empty string.
}

function DUTParamCount : Integer;
{
   A replacement for the VCL's ParamCount that
   does not suffer from built-in buffer size limitations.
   It'll handle whatever the system passes to it.

   It returns the number of parametres in the command line
   (excluding the executable's name and path).
}

function DUTParamStr(Index : Integer) : string;
{
   A replacement for the VCL's ParamStr that
   does not suffer from built-in buffer size limitations.
   It'll handle whatever the system passes to it.

   It returns the Index-th parametre in the command line;
   DUTParamStr(0) returns the executable's name and path.
   If Index > DUTParamCount (or Index < 0), returns an empty string.
}

function DUTParamSwitch(const Switch : String) : Boolean;
{
   Returns True if one of the parametres is the command-line Switch,
   False otherwise.  The comparison is not case-sensitive.
   Example:
   //Test for '/help' or '-help' switch
   if DUTParamSwitch('help') then ...
}

{ #################### Window Utilities #################### }

function GetChildByID(hParent : HWND; SoughtID : LongInt) : HWND;
{
   Retrieves a child's window handle based on its ID.
   Returns zero in case of failure.
hParent
   Handle of the window whose children are to be interrogated.
SoughtID
   The ID of the sought child.
}

function GetShiftState : TShiftState;
{
   Reconstitutes the ShiftState.
}

function ShiftStateToKeys(ShiftState : TShiftState) : Word;
{
   Reverses Forms.KeysToShiftState
   (useful when converting a Delphi message to a Windows message)
   Note that we throw in MK_ALT and MK_DOUBLE, which aren't defined in Windows.pas
}

function ForceForegroundWindow(hWnd : THandle) : Boolean;
{
   Windows 98 has a SetForegroundWindow() problem.
   Under 98, by default SetForegroundWindow() will NOT
   bring an application's window above another application.
   Code supplied by Daniel P. Stasinski <azdse@uswest.net>
}

{ #################### Menu Manipulation #################### }

function GetHighlightedMenuItem(Menu : HMENU) : Integer;
{
   Returns the MF_BYPOSITION index of the currently highlighted menu entry
   or -1 if there is no such entry.
}

function GetMenuItemInfoByCaption(Menu : HMENU; Caption : String; var MII : TMenuItemInfo) : Boolean;
{
   Fills the TMenuItemInfo for a menu item by its caption.
   The search is not case-sensitive and is recursive.
   Returns False if no menu item had the caption sought.
   The MII.dwTypeData member should not be used since
   it'll point to a freed String (it matches Caption anyway).
}

function IsAccel(VK : Word; const Str : string): Boolean;
{
   This function replaces Forms.IsAccel and fixes its shortcoming:
   It will properly ignore double-ampersands.
}

implementation

uses
   Graphics, { TCanvas, TBitmap, TIcon }
   SysUtils; { TMethod, Exception, IntToStr }

type
   //used by GetChildByID
   PChildSought = ^TChildSought;
   TChildSought = packed record
      FoundWnd : HWND;
      SoughtID : LongInt;
   end; { TChildSought }

{ #################### WinNT Security ID (SID) Functions #################### }

function SIDToStr(SID : PSID) : String;
{
   Converts the SID into its string representation.
   An exception is raised if the SID is invalid.
   The string will be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*
   SID_REVISION is a single byte
   SID_IDENTIFIER_AUTHORITY is 0x000000000000
   SIDSubAuthorities are DWords

   We need this function because ConvertSIDToStringSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}
var
   pSIA             : PSIDIdentifierAuthority;
   dwSubAuthorities,
   dwSIDRev,
   dwCounter        : DWORD;
begin
   dwSIDRev := SID_REVISION;
   //Test if SID passed in is valid
   if not IsValidSID(SID) then raise Exception.Create('SIDtoStr: Invalid SID');

   pSIA := GetSIDIdentifierAuthority(SID);
   dwSubAuthorities := GetSIDSubAuthorityCount(SID)^;
   //Prepare S-SID_REVISION-
// Result := Format('S-%u-', [dwSIDRev]);
   Result := Format('S-%d-', [dwSIDRev]);
   //Prepare SID.IdentifierAuthority
   if (pSIA^.Value[0] <> 0) or (pSIA^.Value[1] <> 0) then begin
      { Note: This hex format is undocumented }
//    Result := Result + Format('0x%02x%02x%02x%02x%02x%02x', [
      Result := Result + Format('0x%2x%2x%2x%2x%2x%2x', [
                                 pSIA^.Value[0],
                                 pSIA^.Value[1],
                                 pSIA^.Value[2],
                                 pSIA^.Value[3],
                                 pSIA^.Value[4],
                                 pSIA^.Value[5] ]);
   end else begin
      //Value[0..1] are ignored; should be zero
//    Result := Result + Format('%u', [
      Result := Result + Format('%d', [
                                 DWORD(pSIA^.Value[5]) +
                                 DWORD(pSIA^.Value[4] shl  8) +
                                 DWORD(pSIA^.Value[3] shl 16) +
                                 DWORD(pSIA^.Value[2] shl 24)   ]);
   end; { else }
   //Loop through SID.SubAuthority array
   for dwCounter := 0 to (dwSubAuthorities - 1) do begin
//    Result := Result + Format('-%u', [GetSIDSubAuthority(SID, dwCounter)^]);
      Result := Result + Format('-%d', [GetSIDSubAuthority(SID, dwCounter)^]);
   end; { for }
end; { SIDToStr }
(*
procedure StrToSID(const SIDName : String; SID : PSID; SIDLen : DWORD);
{
   Writes the string SIDName to the already-allocated SID of size SIDLen.
   An exception is raised if any of the arguments is invalid or the conversion fails.
   The SIDName should be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*

   We need this function because ConvertStringSIDToSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}
var
   pSID,
   pSubAuthorityCount : PChar;
   iSubStr,
   iSubStrTerminator  : Integer;
   ValueStr           : String;
   pSIA               : PSIDIdentifierAuthority;
   i,
   AuthorityCount     : DWORD;
begin
{
   //SID
   type SID = record
      Revision,
      SubAuthorityCount   : BYTE;
      IdentifierAuthority : TSIDIdentifierAuthority;
      //SID will be shorter if SubAuthority isn't completely filled in
//    SubAuthority        : Array[1..SubAuthorityCount] of DWORD;
      SubAuthority        : Array[1..SID_MAX_SUB_AUTHORITIES] of DWORD;
   end;
   //SID_IDENTIFIER_AUTHORITY
   TSIDIdentifierAuthority = record
      Value: array[0..5] of Byte;
   end;
}
   //SIDLen should be at least big enough for zero sub-authorities
// if SIDLen < (SizeOf(_SID_IDENTIFIER_AUTHORITY) + 2) then
   if SIDLen < (SizeOf(TSIDIdentifierAuthority) + 2*SizeOf(BYTE)) then
      raise Exception.Create ('SIDToStr: Bad SIDLen');

   //SIDName should start with 'S-'
   ValueStr := Copy(SIDName, 1, 2);
   if ValueStr <> 'S-' then raise Exception.Create('SIDToStr: Bad SIDName');

   iSubStr := 3;
   //iSubStr points to the beginning of the SID_REVISION substring
   iSubStrTerminator := iSubStr + Pos('-', Copy(SIDName, iSubStr, 9999)) - 1;
   if iSubStrTerminator < iSubStr then raise Exception.Create('SIDToStr: Bad SIDName');
   //iSubStrTerminator points to the '-' after the SID_REVISION substring
   ValueStr := Copy(SIDName, iSubStr, iSubStrTerminator - iSubStr);

   pSID := PChar(SID);
   //Write the SID_REVISION byte
   pSID^ := Char(StrToInt(ValueStr));
   Inc(pSID);
   pSubAuthorityCount := pSID;
   //pSubAuthorityCount points to the SubAuthorityCount byte
   Inc(pSID);
   //pSID points to the IdentifierAuthority field

   iSubStr := iSubStrTerminator + 1;
   //iSubStr now points to the beginning of the IdentifierAuthority substring
   iSubStrTerminator := iSubStr + Pos('-', Copy(SIDName, iSubStr, 9999)) - 1;
   if iSubStrTerminator < iSubStr then raise Exception.Create('SIDToStr: Bad SID');
   //iSubStrTerminator now points to the '-' after the IdentifierAuthority substring

   ValueStr := Copy(SIDName, iSubStr, iSubStrTerminator - iSubStr);
   if Length(ValueStr) < 1 then raise Exception.Create ('SIDToStr: Bad SID');

   pSIA := PSIDIdentifierAuthority(pSID);
   Inc(pSID, SizeOf(TSIDIdentifierAuthority));
   //pSID now points to the beginning of the SubAuthority array
   Dec(SIDLen, 2 + SizeOf(TSIDIdentifierAuthority));
   //SIDLen is now the SubAuthority array size

   //Fill in the SID's IdentifierAuthority
   if ValueStr[1] = '0x' then begin { Colin had this as 'x'?? }
      { Note: This hex format is undocumented }
      if Length(ValueStr) <> 14 then raise Exception.Create('SIDToStr: Bad SID');
      pSIA^.Value[0] := StrToInt('$' + Copy(ValueStr,  3, 2));
      pSIA^.Value[1] := StrToInt('$' + Copy(ValueStr,  5, 2));
      pSIA^.Value[2] := StrToInt('$' + Copy(ValueStr,  7, 2));
      pSIA^.Value[3] := StrToInt('$' + Copy(ValueStr,  9, 2));
      pSIA^.Value[4] := StrToInt('$' + Copy(ValueStr, 11, 2));
      pSIA^.Value[5] := StrToInt('$' + Copy(ValueStr, 13, 2))
   end else begin
      i := StrToInt(ValueStr);
      pSIA^.Value[0] := 0;
      pSIA^.Value[1] := 0;
      pSIA^.Value[2] := (i shl 24) and $FF;
      pSIA^.Value[3] := (i shl 16) and $FF;
      pSIA^.Value[4] := (i shl  8) and $FF;
      pSIA^.Value[5] :=  i         and $FF;
   end; { else }

   iSubStr := iSubStrTerminator + 1;
   ValueStr := Copy(SIDName, iSubStr, 9999);
   //ValueStr is the SubAuthority[AuthorityCount] substring
   AuthorityCount := 0;
   //iSubStr now points to the SubAuthority[AuthorityCount] substring
   while ValueStr <> '' do begin
      iSubStrTerminator := Pos('-', ValueStr);
      if iSubStrTerminator <= 0 then begin
         i := StrToInt(ValueStr);
         ValueStr := '';
      end else begin
         //iSubStrTerminator now points to the '-' that follows the SubAuthority[AuthorityCount] substring
         i := StrToInt(Copy(ValueStr, 1, iSubStrTerminator - 1));
         ValueStr := Copy(ValueStr, iSubStrTerminator + 1, 9999);
      end; { else }
      if SIDLen < SizeOf(DWORD) then raise Exception.Create('SIDToStr: Bad SIDLen');
      PDWORD(pSID)^ := i;
      Inc(pSID, SizeOf(DWORD));
      //pSID now points to SubAuthority[AuthorityCount + 1]
      Dec(SIDLen, SizeOf(DWORD));
      //SIDLen is now the size of SubAuthority[(AuthorityCount + 1)..]
      Inc(AuthorityCount);
   end; { while }
   //Write the SubAuthorityCount byte
   pSubAuthorityCount^ := Char(AuthorityCount);

   //Check the resulting SID
   if not IsValidSID(SID) then raise Exception.Create('SIDToStr: Bad SID');
end; { StrToSID }
*)
function StrToSID(const SIDName : String) : PSID;
{
   Allocates and fills in an SID based on the string SIDName.
   An exception is raised (and no memory allocated) if the
   SID creation fails.
   The SIDName should be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*
   The caller is responsible for calling FreeSID to free the PSID.

   We need this function because ConvertStringSIDToSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}
var
   iSubStr,
   iSubStrTerminator  : Integer;
   ValueStr           : String;
   SIA                : TSIDIdentifierAuthority;
   i,
   AuthorityCount     : DWORD;
   SubAuthority       : Array[1..SID_MAX_SUB_AUTHORITIES] of DWORD;
begin
{
   //SID
   type SID = record
      Revision,
      SubAuthorityCount   : BYTE;
      IdentifierAuthority : TSIDIdentifierAuthority;
      //SID will be shorter if SubAuthority isn't completely filled in
//    SubAuthority        : Array[1..SubAuthorityCount] of DWORD;
      SubAuthority        : Array[1..SID_MAX_SUB_AUTHORITIES] of DWORD;
   end;
   //SID_IDENTIFIER_AUTHORITY
   TSIDIdentifierAuthority = record
      Value: array[0..5] of Byte;
   end;
}
   //SIDName should start with 'S-'
   ValueStr := Copy(SIDName, 1, 2);
   if ValueStr <> 'S-' then raise Exception.Create('SIDToStr: SIDName must begin with "S-"');

   iSubStr := 3;
   //iSubStr points to the beginning of the SID_REVISION substring
   iSubStrTerminator := iSubStr + Pos('-', Copy(SIDName, iSubStr, 9999)) - 1;
   if iSubStrTerminator < iSubStr then raise Exception.Create('SIDToStr: Bad SIDName: Unable to recognise SID_REVISION substring');
   //iSubStrTerminator points to the '-' after the SID_REVISION substring
// ValueStr := Copy(SIDName, iSubStr, iSubStrTerminator - iSubStr);

   iSubStr := iSubStrTerminator + 1;
   //iSubStr now points to the beginning of the IdentifierAuthority substring
   iSubStrTerminator := iSubStr + Pos('-', Copy(SIDName, iSubStr, 9999)) - 1;
   if iSubStrTerminator < iSubStr then raise Exception.Create('SIDToStr: Bad SIDName: Unable to recognise SID_IDENTIFIER_AUTHORITY substring');
   //iSubStrTerminator now points to the '-' after the IdentifierAuthority substring
   ValueStr := Copy(SIDName, iSubStr, iSubStrTerminator - iSubStr);

   //Fill in the SID's IdentifierAuthority
   if ValueStr[1] = '0x' then begin
      { Note: This hex format is undocumented }
      if Length(ValueStr) <> 14 then raise Exception.Create('SIDToStr: Bad SIDName: SID_IDENTIFIER_AUTHORITY incorrectly formatted');
      SIA.Value[0] := StrToInt('$' + Copy(ValueStr,  3, 2));
      SIA.Value[1] := StrToInt('$' + Copy(ValueStr,  5, 2));
      SIA.Value[2] := StrToInt('$' + Copy(ValueStr,  7, 2));
      SIA.Value[3] := StrToInt('$' + Copy(ValueStr,  9, 2));
      SIA.Value[4] := StrToInt('$' + Copy(ValueStr, 11, 2));
      SIA.Value[5] := StrToInt('$' + Copy(ValueStr, 13, 2))
   end else begin
      i := StrToInt(ValueStr);
      SIA.Value[0] := 0;
      SIA.Value[1] := 0;
      SIA.Value[2] := (i shr 24) and $FF;
      SIA.Value[3] := (i shr 16) and $FF;
      SIA.Value[4] := (i shr  8) and $FF;
      SIA.Value[5] :=  i         and $FF;
   end; { else }

   iSubStr := iSubStrTerminator + 1;
   ValueStr := Copy(SIDName, iSubStr, 9999);
   //ValueStr is the SubAuthority[AuthorityCount] substring
   AuthorityCount := 0;
   //iSubStr now points to the SubAuthority[AuthorityCount] substring
   while ValueStr <> '' do begin
      Inc(AuthorityCount);
      iSubStrTerminator := Pos('-', ValueStr);
      if iSubStrTerminator <= 0 then begin
         SubAuthority[AuthorityCount] := StrToInt(ValueStr);
         ValueStr := '';
      end else begin
         //iSubStrTerminator now points to the '-' that follows the SubAuthority[AuthorityCount] substring
         SubAuthority[AuthorityCount] := StrToInt(Copy(ValueStr, 1, iSubStrTerminator - 1));
         ValueStr := Copy(ValueStr, iSubStrTerminator + 1, 9999);
      end; { else }
   end; { while }

   if not AllocateAndInitializeSID(
      SIA,
      AuthorityCount,
      SubAuthority[1],
      SubAuthority[2],
      SubAuthority[3],
      SubAuthority[4],
      SubAuthority[5],
      SubAuthority[6],
      SubAuthority[7],
      SubAuthority[8],
      Result
   ) then raise Exception.Create('SIDToStr: Failed to create the SID');

   //Note that there is no guarantee the resulting SID has the same
   //Revision substring as the input SIDName.
end; { StrToSID }

{ #################### Message Utilities #################### }

function FlashMessagePanel(const Msg : String; AType : TMsgDlgType; HelpCtx : Longint; X, Y : Integer; const HelpFileName : String; MilliSeconds : Integer) : TForm;
{
   Flashes a message panel (a button-less dialog) in non-modal fashion.
   Msg          is the message displayed on the panel
   AType        is the TMsgDlgType (as per the MessageDlg functions)
   HelpCtx      is an optional Help context (zero for none)
   X, Y         are optional Left and Top positions (-1 for default)
   HelpFileName is an optional Help file name ('' for default)
   Milliseconds is how long to display the panel

   If Milliseconds is non-zero, FlashMessagePanel waits for the specified time and then
   disposes of the message panel; the function returns Nil.  The window will have no
   BorderIcons in that case (no close box).
   If Milliseconds is zero (or less), FlashMessagePanel throws the message panel up
   and then returns its TForm reference.  It is up to the caller to Free it.  The
   window will have a close box.
}
const
   mcVertSpacing = 10;
   mcButtonHeight = 14;
var
   DialogUnits  : TPoint;
   VertSpacing,
   ButtonHeight : Integer;

   function GetAveCharSize(Canvas : TCanvas) : TPoint;
   {
      Returns the average width and maximum height
      of the A..Z and a..z characters in the specified Canvas.
   }
   var
      I      : Integer;
      Buffer : array[0..51] of Char;
   begin
      for I := 0 to 25 do Buffer[I] := Chr(I + Ord('A')); //A..Z
      for I := 0 to 25 do Buffer[I + 26] := Chr(I + Ord('a')); //a..z
      GetTextExtentPoint(Canvas.Handle, Buffer, 52, TSize(Result));
      Result.X := Result.X div 52;
   end; { GetAveCharSize }

begin
   Result := CreateMessageDialog(Msg, AType, []);
   with Result do begin
      try
         HelpContext := HelpCtx;
         HelpFile := HelpFileName;
         if X >= 0 then Left := X;
         if Y >= 0 then Top := Y;
         //Correct ClientHeight for the lack of buttons
//       ClientHeight := IconTextHeight + ButtonHeight + VertSpacing + VertMargin * 2;
         DialogUnits := GetAveCharSize(Canvas);
         VertSpacing := MulDiv(mcVertSpacing, DialogUnits.Y, 8);
         ButtonHeight := MulDiv(mcButtonHeight, DialogUnits.Y, 8);
         ClientHeight := ClientHeight - (ButtonHeight + VertSpacing);
         FormStyle := fsStayOnTop;
         if Milliseconds > 0 then BorderIcons := [];

         Show;
         Refresh;
         if Milliseconds <= 0 then Exit;

         Sleep(Milliseconds);
         Hide;
         Free;
      except
         Free;
      end; { try }
   end; { with }
   Result := Nil;
end; { FlashMessagePanel }

{ #################### Spawning Utilities #################### }

function RunItAndWait(AppCommand : String) : LongInt;
{
   "Shells out" to the AppCommand (which must be a DOS or Windows command line)
   and waits for it to terminate.
   The exit code (or an error code) is returned.
   For the 16-bit Setup.exe stub app used by Install Shield Express,
   this won't work.  I'm at a loss to explain why.
}
var
   si : TStartupInfo;
   pi : TProcessInformation;
begin
   FillChar(si, SizeOf(si), #0);
   FillChar(pi, SizeOf(pi), #0);
   si.cb := SizeOf(si);
   si.wShowWindow := SW_SHOWNORMAL;
   if not CreateProcess(Nil, PChar(AppCommand), Nil, Nil, False, 0, Nil, Nil, si, pi) then begin
      Result := GetLastError;
      Exit;
   end; { then }
(*
   if -1 = WaitForInputIdle(pi.hProcess, INFINITE) then begin
      Result := GetLastError;
      Exit;
   end; { then }
*)
// WaitForSingleObject(pi.hProcess, INFINITE); //uses Windows
// GetExitCodeProcess(pi.hProcess, Result);
   repeat
      //Allow application to refresh, move window, etc.
      Application.ProcessMessages;
      //Half-second cycle
      if (WAIT_FAILED = WaitForSingleObject(pi.hProcess, 500)) or
         (not GetExitCodeProcess(pi.hProcess, Result)) then begin
         Result := GetLastError;
         Exit;
      end; { then }
   until (Result <> STILL_ACTIVE);
{
   WaitForSingleObject returns when the specified object is in the signaled state
   (or when the time-out interval elapses)
   When the Object to wait for is a Process,
   its handle is returned by the CreateProcess or OpenProcess function.
   A process object's state is signaled when the process terminates.
}
end; { RunItAndWait }

{ #################### System Info #################### }

function DUTGetUserName : String;
{
   Obtains the User Name.
   Returns the '.Default' string if no "user" has logged on.
}
var
   Size : DWORD;
begin
   Size := MAX_PATH;
   Result := StringOfChar(#0, Size);
   if Windows.GetUserName(PChar(Result), Size) then begin
      SetLength(Result, Size - 1);
   end else begin
      Result := '.Default';
   end; { else }
end; { DUTGetUserName }

function DUTGetNetworkUserName : String;
{
   Obtains the network User Name.
   Returns an empty string if there is no network or
   the user did not log on.
   Raises an exception otherwise.
}
var
   Size : DWORD;
begin
   Size := MAX_PATH;
   while True do begin
      Result := StringOfChar(#0, Size);
      SetLastError(0);
      if NO_ERROR = WNetGetUser(Nil, PChar(Result), Size) then begin
         SetLength(Result, Size - 1);
         Exit;
      end; { then }
      case GetLastError of
      ERROR_MORE_DATA :
         //Buffer was too small
         Size := 2*Size;
      ERROR_NO_NETWORK,
      ERROR_NOT_LOGGED_ON : begin
         //No network is present or the user did not log on
         Result := '';
         Exit;
      end; { ERROR_NO_NETWORK, ERROR_NOT_LOGGED_ON }
      else
//      ERROR_NOT_CONNECTED :
//      ERROR_EXTENDED_ERROR :
//      ERROR_NO_NET_OR_BAD_PATH :
         raise Exception.CreateFmt('WNetGetUser failed.  Error #%d: %s', [GetLastError, SysErrorMessage(GetLastError)]);
      end; { case }
   end; { while }
end; { DUTGetNetworkUserName }

function IsWindows95 : Boolean;
{
   Returns True if the OS is Windows 95 (or 98)
}
var
   VersionInfo : TOSVersionInfo;
begin
   // set the size of the record
   VersionInfo.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
   if {Windows.}GetVersionEx(VersionInfo) then begin
      Result := VersionInfo.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS;
      {
      dwPlatformID could also be
      VER_PLATFORM_WIN32s or //Win32s
      VER_PLATFORM_WIN32_NT  //Windows NT
      }
   end else begin
      raise Exception.Create('IsWindows95: Unable to GetVersionEx');
   end; { else }
end; { IsWindows95 }

function IsWindowsNT : Boolean;
{
   Returns True if the OS is Windows NT
}
var
   VersionInfo : TOSVersionInfo;
begin
   // set the size of the record
   VersionInfo.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
   if {Windows.}GetVersionEx(VersionInfo) then begin
      Result := VersionInfo.dwPlatformId = VER_PLATFORM_WIN32_NT;
      {
      dwPlatformID could also be
      VER_PLATFORM_WIN32s or     //Win32s
      VER_PLATFORM_WIN32_WINDOWS //Windows 95
      }
   end else begin
      raise Exception.Create('IsWindowsNT: Unable to GetVersionEx');
   end; { else }
end; { IsWindowsNT }

function GetLocaleNumberFmt(Locale : LCID; var NumSepStr : String; var ThsndsStr : String) : TNumberFmt;
{
   Obtains the TNumberFmt record from the system for the specified Locale;
   NumSepStr and ThsndsStr must be supplied as they'll be pointed to by
   the TNumberFmt.
}
var
   NumberStr : Array[0..255] of Char;
begin
   { Number of decimals }
   GetLocaleInfo(Locale, LOCALE_IDIGITS, @NumberStr, 256); { Max length 3 (includes closing #0) }
   Result.NumDigits := StrToInt(String(NumberStr));
   { Leading zero in decimal fields? }
   GetLocaleInfo(Locale, LOCALE_ILZERO, @NumberStr, 256); { Max length 2 }
   Result.LeadingZero := StrToInt(String(NumberStr)); { 0..1 }
   { Group size left of decimal }
   { SGrouping is a series of '%d;' values closing with '0' -- use the last one }
   GetLocaleInfo(Locale, LOCALE_SGROUPING, @NumberStr, 256); { Unspecified max length }
   Result.Grouping := StrToInt(String(NumberStr[LastDelimiter(';', String(NumberStr)) - 2]));
   { Pointer to decimal separator string }
   GetLocaleInfo(Locale, LOCALE_SDECIMAL, @NumSepStr, 2); { Max length 2 }
   Result.lpDecimalSep := @(NumSepStr);
   { Pointer to thousand separator string }
   GetLocaleInfo(Locale, LOCALE_STHOUSAND, @ThsndsStr, 2); { Max length 2 }
   Result.lpThousandSep := @(ThsndsStr);
   { Negative number ordering }
   GetLocaleInfo(Locale, LOCALE_INEGNUMBER, @NumberStr, 256); { Max length 2 }
   Result.NegativeOrder := StrToInt(String(NumberStr));
end; { GetLocaleNumberFmt }

function GetLocaleCurrencyFmt(Locale : LCID; var NumSepStr : String; var ThsndsStr : String; var CrrncyStr : String) : TCurrencyFmt;
{
   Obtains the TCurrencyFmt record from the system for the specified Locale;
   NumSepStr, ThsndsStr and CrrncyStr must be supplied as they'll be pointed
   to by the TCurrencyFmt.
}
var
   NumberStr : Array[0..255] of Char;
begin
   { Number of decimals }
   GetLocaleInfo(Locale, LOCALE_ICURRDIGITS, @NumberStr, 256); { Max length 3 (includes closing #0) }
   Result.NumDigits := StrToInt(String(NumberStr));
   { Leading zero in decimal fields? }
   GetLocaleInfo(Locale, LOCALE_ILZERO, @NumberStr, 256); { Max length 2 }
   Result.LeadingZero := StrToInt(String(NumberStr)); { 0..1 }
   { Group size left of decimal }
   { SGrouping is a series of '%d;' values closing with '0' -- use the last one }
   GetLocaleInfo(Locale, LOCALE_SMONGROUPING, @NumberStr, 256); { Unspecified max length }
   Result.Grouping := StrToInt(String(NumberStr[LastDelimiter(';', String(NumberStr)) - 2]));
   { Pointer to decimal monetary separator string }
   GetLocaleInfo(Locale, LOCALE_SMONDECIMALSEP, @NumSepStr, 2); { Max length 2 }
   Result.lpDecimalSep := @(NumSepStr);
   { Pointer to thousand monetary separator string }
   GetLocaleInfo(Locale, LOCALE_SMONTHOUSANDSEP, @ThsndsStr, 2); { Max length 2 }
   Result.lpThousandSep := @(ThsndsStr);
   { Negative currency ordering }
   GetLocaleInfo(Locale, LOCALE_INEGCURR, @NumberStr, 256); { Max length 3 }
   Result.NegativeOrder := StrToInt(String(NumberStr)); { 0..15 }
   { Positive currency ordering }
   GetLocaleInfo(Locale, LOCALE_ICURRENCY, @NumberStr, 256); { Max length 2 }
   Result.PositiveOrder := StrToInt(String(NumberStr)); { 0..3 }
   { Pointer to currency symbol string }
   GetLocaleInfo(Locale, LOCALE_SCURRENCY, @CrrncyStr, 2); { Max length unspecified }
   Result.lpCurrencySymbol := @(CrrncyStr);
end; { GetLocaleCurrencyFmt }

function GetDOSEnvironmentVar(const VarName : String) : String;
{
   Returns the DOS Environment variable VarName (such as PATH or TEMP).
   If VarName isn't defined, returns an empty string.
   For some unexplained reason, although GetEnvironmentStrings returns
   a block that DOES contain a given VarName, GetEnvironmentVariable
   may fail to find it.  This problem vanishes if one happens to be
   debugging from the Delphi IDE --that is, if you put a breakpoint in
   your code, EVEN AFTER THE CALL TO THIS FUNCTION, GetEnvironmentVariable
   works like it should.  I'm baffled.
}
var
   Len : Integer;
begin
   Result := '';
   //Returns the length of the buffer required (including terminating null)
   Len := GetEnvironmentVariable(PChar(VarName), PChar(Result), 0);
   //Len = 0 means no such environment variable
   if Len = 0 then Exit;
   Result := StringOfChar(#0, Len);
   //Should return Len - 1 (it excludes the terminating null)
   GetEnvironmentVariable(PChar(VarName), PChar(Result), Len);
   Result := TrimRight(Result);
end; { GetDOSEnvironmentVar }

function DUTGetParamStr(P : PChar; var Param : String) : PChar;
{
   Private function that supports DUTParamCount and DUTParamStr.
   Scans the input string P and
   sets Param to the first parametre (ignoring any enclosing quotes) and
   returns a pointer to the first character after the parametre
   (also skipping any enclosing quotes).
}
begin
   while True do begin
      //Skip blanks, control characters but not concluding zero
      while (P[0] <> #0) and (P[0] <= ' ') do Inc(P);
      if not ((P[0] = '"') and (P[1] = '"')) then break;
      //Skip empty parametre
      Inc(P, 2)
   end; { while }
   //P points at the beginning of the parametre (or at the first enclosing quote)
   Param := '';
   if P[0] = '"' then begin
      //Quoted parametre
      //Skip the quote
      Inc(P);
      //Accumulate characters up to but excluding a zero or closing quote
      //We accept control characters?!?
      while (P[0] <> #0) and (P[0] <> '"') do begin
         Param := Param + String(P[0]);
         Inc(P);
      end; { while }
      //Skip closing quote
      if P[0] <> #0 then Inc(P);
   end else begin
      //Unquoted parametre
      //Accumulate characters up to but excluding next control character or space
      while P[0] > ' ' do begin
         Param := Param + String(P[0]);
         Inc(P);
      end; { while }
   end; { else }
   //P points after the parametre, skipping enclosing quotes
   Result := P;
end; { DUTGetParamStr }

function DUTParamCount : Integer;
{
   A replacement for the VCL's ParamCount that
   does not suffer from built-in buffer size limitations.
   It'll handle whatever the system passes to it.

   It returns the number of parametres in the command line
   (excluding the executable's name and path).
}
var
   P : PChar;
   S : String;
begin
   Result := -1;
   //Strip out the executable's name and path
   P := DUTGetParamStr(GetCommandLine, S);
   repeat
      //Obtain a parametre
      P := DUTGetParamStr(P, S);
      Inc(Result);
   until S = '';
end; { DUTParamCount }

function DUTParamStr(Index : Integer) : string;
{
   A replacement for the VCL's ParamStr that
   does not suffer from built-in buffer size limitations.
   It'll handle whatever the system passes to it.

   It returns the Index-th parametre in the command line;
   DUTParamStr(0) returns the executable's name and path.
   If Index > DUTParamCount (or Index < 0), returns an empty string.
}
var
   nChars  : Integer;
   P       : PChar;
begin
   if Index = 0 then begin
      //Must return the executable's name and path
      nChars := MAX_PATH;
      while True do begin
         Result := StringOfChar(#0, nChars);
         //GetModuleFileName truncates if the result is too long
         if nChars > GetModuleFileName(0, PChar(Result), nChars) then break;
         nChars := 2*nChars;
      end; { while }
      //Adjust length of Result
      SetLength(Result, Pos(#0, Result + #0) - 1);
   end else begin
      Result := '';
      if Index < 0 then Exit;
      P := GetCommandLine;
      while True do begin
         P := DUTGetParamStr(P, Result);
         if (Index <= 0) or (Result = '') then Exit;
         Dec(Index);
      end; { while }
   end; { else }
end; { DUTParamStr }

function DUTParamSwitch(const Switch : String) : Boolean;
{
   Returns True if one of the parametres is the command-line Switch,
   False otherwise.  The comparison is not case-sensitive.
   Example:
   //Test for '/help' or '-help' switch
   if DUTParamSwitch('help') then ...
}
var
   n,
   i : Integer;
   S : string;
begin
   Result := False;
   i := 1;
   n := DUTParamCount;
   while (not Result) and (i <= n) do begin
      S := DUTParamStr(i);
      Result := (S[1] in ['-', '/']) and
                (0 = ANSICompareText(Copy(S, 2, MAXINT), Switch));
      Inc(i);
   end; { while }
end; { DUTParamSwitch }

{ #################### Window Utilities #################### }

function GetChildByID(hParent : HWND; SoughtID : LongInt) : HWND;
{
   This function uses EnumChildWindows to retrieve a child's handle based on its ID.
   Returns zero in case of failure.
hParent
   Handle of the window whose children are to be interrogated.
SoughtID
   The ID of the sought child.
}
   // Note that BOOL <> Boolean type.  Important!
   function EnumChildWndProc(Child : HWND; Data : LParam) : BOOL; stdcall;
   {
      Callback used by EnumChildWindows to locate a specific child window of a dialog.
   }
   begin
      Result := True;
      if PChildSought(Data)^.SoughtID <> GetWindowLong(Child, GWL_ID) then Exit;
      PChildSought(Data)^.FoundWnd := Child;
      Result := False;
   end; { EnumChildWndProc }

var
   ChildWnd : TChildSought;
begin
   ChildWnd.FoundWnd := 0;
   ChildWnd.SoughtID := SoughtID;
   EnumChildWindows(hParent, @EnumChildWndProc, {LPARAM}LongInt(@ChildWnd));
   Result := ChildWnd.FoundWnd;
end; { GetChildByID }

function GetShiftState : TShiftState;
{
   Reconstitutes the ShiftState.
}
begin
   Result := [];
   if GetAsyncKeyState(VK_MENU)    < 0 then Include(Result, ssAlt);
   if GetAsyncKeyState(VK_LBUTTON) < 0 then if 0 <> GetSystemMetrics(SM_SWAPBUTTON) then Include(Result, ssRight) else Include(Result, ssLeft);
   if GetAsyncKeyState(VK_MBUTTON) < 0 then Include(Result, ssMiddle);
   if GetAsyncKeyState(VK_RBUTTON) < 0 then if 0 <> GetSystemMetrics(SM_SWAPBUTTON) then Include(Result, ssLeft) else Include(Result, ssRight);
   if GetAsyncKeyState(VK_SHIFT)   < 0 then Include(Result, ssShift);
   if GetAsyncKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
end; { GetShiftState }

function ShiftStateToKeys(ShiftState : TShiftState) : Word;
{
   Reverses Forms.KeysToShiftState
   (useful when converting a Delphi message to a Windows message)
   Note that we throw in MK_ALT and MK_DOUBLE, which aren't defined in Windows.pas
}
begin
  Result := 0;
  if ssShift  in ShiftState then Result := Result or MK_SHIFT;
  if ssCtrl   in ShiftState then Result := Result or MK_CONTROL;
  if ssLeft   in ShiftState then Result := Result or MK_LBUTTON;
  if ssRight  in ShiftState then Result := Result or MK_RBUTTON;
  if ssMiddle in ShiftState then Result := Result or MK_MBUTTON;
  if ssAlt    in ShiftState then Result := Result or MK_ALT;
  if ssDouble in ShiftState then Result := Result or MK_DOUBLE;
end; { ShiftStateToKeys }

function ForceForegroundWindow(hWnd : THandle) : Boolean;
{
   Windows 98 has a SetForegroundWindow() problem.
   Under 98, by default SetForegroundWindow() will NOT
   bring an application's window above another application.
   Code supplied by Daniel P. Stasinski <azdse@uswest.net>
}
const
   SPI_GETFOREGROUNDLOCKTIMEOUT = $2000;
   SPI_SETFOREGROUNDLOCKTIMEOUT = $2001;
var
   TimeOut : DWORD;
begin
   //Problem occurs with WinNT >= 5 and Win9x > 4.0
   if (
       (Win32Platform = VER_PLATFORM_WIN32_NT)
       and
       (Win32MajorVersion > 4)
      )
      or
      (
       (Win32Platform = VER_PLATFORM_WIN32_WINDOWS)
       and
       (
        (Win32MajorVersion > 4)
        or
        (
         (Win32MajorVersion = 4)
         and
         (Win32MinorVersion > 0)
        )
       )
      )
   then begin
      //Save old ForegroundLockTimeOut setting
      SystemParametersInfo(SPI_GETFOREGROUNDLOCKTIMEOUT, 0, @TimeOut, 0);
      //Change it temporarily to null
      SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, TObject(0), SPIF_SENDCHANGE);
      //Bring the window to the foreground
      Result := SetForegroundWindow(hWnd);
      //Restore ForegroundLockTimeOut setting
      SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, TObject(TimeOut), SPIF_SENDCHANGE);
   end else begin
      Result := SetForegroundWindow(hWnd);
   end; { else }
end; { ForceForegroundWindow }

{ #################### Menu Manipulation #################### }

function GetHighlightedMenuItem(Menu : HMENU) : Integer;
{
   Returns the MF_BYPOSITION index of the currently highlighted menu entry
   or -1 if there is no such entry.
}
var
   MenuIndex    : Integer;
   MenuItemInfo : TMenuItemInfo;
begin
   Result := -1;
   MenuItemInfo.cbSize := SizeOf(MenuItemInfo);
   MenuItemInfo.fMask := MIIM_STATE;
   for MenuIndex := 0 to (GetMenuItemCount(Menu) - 1) do begin
      GetMenuItemInfo(Menu, MenuIndex, True {MF_BYPOSITION}, MenuItemInfo);
      if (MenuItemInfo.fState and MFS_HILITE) = MFS_HILITE then begin
         Result := MenuIndex;
         break;
      end; { then }
   end; { for }
end; { GetHighlightedMenuItem }

function GetMenuItemInfoByCaption(Menu : HMENU; Caption : String; var MII : TMenuItemInfo) : Boolean;
{
   Fills the TMenuItemInfo for a menu item by its caption.
   The search is not case-sensitive and is recursive.
   Returns False if no menu item had the caption sought.
   The MII.dwTypeData member should not be used since
   it'll point to a freed String (it matches Caption anyway).
}
var
   i        : Integer;
   ACaption : String;
begin
   Result := False;
   for i := 0 to (GetMenuItemCount(Menu) - 1) do begin
      FillChar(MII, SizeOf(TMenuItemInfo), #0);
      MII.cbSize := SizeOf(TMenuItemInfo);
      //Retrieve everything
      MII.fMask := MIIM_CHECKMARKS or MIIM_DATA or MIIM_ID or MIIM_STATE or MIIM_SUBMENU or MIIM_TYPE;
      MII.cch := 255;
      ACaption := StringOfChar(#0, MII.cch);
      MII.dwTypeData := PChar(ACaption);
      if not GetMenuItemInfo(Menu, i, True, MII) then raise Exception.Create('GetMenuItemInfoByCaption: Unable to GetMenuItemInfo for item number ' + IntToStr(i));
      //Ignore Separators and Bitmaps
      if (MII.fType = MFT_STRING) and
         (0 = ANSICompareText(Caption, String(MII.dwTypeData)))
      then begin
         Result := True;
      end else if MII.hSubMenu <> 0 then begin
         //Recurse
         Result := GetMenuItemInfoByCaption(MII.hSubMenu, Caption, MII);
      end; { else if }
      if Result then Exit;
   end; { for }
end; { GetMenuItemInfoByCaption }

function IsAccel(VK : Word; const Str : string): Boolean;
{
   This function replaces Forms.IsAccel and fixes its shortcoming:
   It will properly ignore double-ampersands.
}
var
   Limit,
   P : Integer;
begin
   {
   Note that if the menu has a shortcut (appended to the caption as Caption + #9 + Shortcut),
   a caption that ENDS with an '&' will activate the menu item on 'Alt-Tab'!
   Note also that in the case of multiple accelerators the LAST one is displayed
   but the FIRST one is the one Windows looks for...
   }
   if Pos(#9, Str) > 0 then begin
      Limit := Pos(#9, Str);
   end else begin
      Limit := Length(Str);
   end; { else }
   P := 1;
   repeat
      P := (P - 1) + Pos('&', Copy(Str, P, Limit - (P - 1)));
   until (P = 0) or (P = Limit) or ('&' <> Str[P + 1]);
   Result := (P <> 0) and (P < Limit) and (0 = AnsiCompareText(Str[P + 1], Char(VK)));
end; { IsAccel }

end. { DUTUtilsSys }
